/*
	G. Wiesenekker. ZZZZZZ a chess program.
	Copyright (C) 1993  G. Wiesenekker
	E-mail: wiesenecker@sara.nl

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "zzzzzz.h"

int DepthMax = 32;
int TimeLimit = 30;
int Append = 1;
int UseBook = 1;
int GameTimeUsed = 0;

/*
I know about cylinder boards, but they give me a headache!
I use a 12x12 board, (8x8 with a double boundary).
*/

#define BOARD_MAX 144

enum {A8 =  26, B8, C8, D8, E8, F8, G8, H8};
enum {A7 =  38, B7, C7, D7, E7, F7, G7, H7};
enum {A6 =  50, B6, C6, D6, E6, F6, G6, H6};
enum {A5 =  62, B5, C5, D5, E5, F5, G5, H5};
enum {A4 =  74, B4, C4, D4, E4, F4, G4, H4};
enum {A3 =  86, B3, C3, D3, E3, F3, G3, H3};
enum {A2 =  98, B2, C2, D2, E2, F2, G2, H2};
enum {A1 = 110, B1, C1, D1, E1, F1, G1, H1};

/*
Sometimes we have to convert from the 'normal' 
index (0..63) to the 'true' index.
*/

int map[64] =
{
	 A8, B8, C8, D8, E8, F8, G8, H8,
	 A7, B7, C7, D7, E7, F7, G7, H7,
	 A6, B6, C6, D6, E6, F6, G6, H6,
	 A5, B5, C5, D5, E5, F5, G5, H5,
	 A4, B4, C4, D4, E4, F4, G4, H4,
	 A3, B3, C3, D3, E3, F3, G3, H3,
	 A2, B2, C2, D2, E2, F2, G2, H2,
	 A1, B1, C1, D1, E1, F1, G1, H1
};

/*
mirror_map comes in handy when we have to initialize
tables which are different for white and black.
*/

int mirror_map[64] =
{
	 A1, B1, C1, D1, E1, F1, G1, H1,
	 A2, B2, C2, D2, E2, F2, G2, H2,
	 A3, B3, C3, D3, E3, F3, G3, H3,
	 A4, B4, C4, D4, E4, F4, G4, H4,
	 A5, B5, C5, D5, E5, F5, G5, H5,
	 A6, B6, C6, D6, E6, F6, G6, H6,
	 A7, B7, C7, D7, E7, F7, G7, H7,
	 A8, B8, C8, D8, E8, F8, G8, H8
};

/* 
This needs no comment
*/

char *nota_table[64] = 
{
	"a8", "b8", "c8", "d8", "e8", "f8", "g8", "h8",
	"a7", "b7", "c7", "d7", "e7", "f7", "g7", "h7",
	"a6", "b6", "c6", "d6", "e6", "f6", "g6", "h6",
	"a5", "b5", "c5", "d5", "e5", "f5", "g5", "h5",
	"a4", "b4", "c4", "d4", "e4", "f4", "g4", "h4",
	"a3", "b3", "c3", "d3", "e3", "f3", "g3", "h3",
	"a2", "b2", "c2", "d2", "e2", "f2", "g2", "h2",
	"a1", "b1", "c1", "d1", "e1", "f1", "g1", "h1"
};
char *nota_inval = "??";
char **nota;

#define GET_NOTA(X) nota[X]

/*
the row numbers used by the program. 
*/

byte_t row_table[64] =
{
	7, 7, 7, 7, 7, 7, 7, 7,
	6, 6, 6, 6, 6, 6, 6, 6,
	5, 5, 5, 5, 5, 5, 5, 5,
	4, 4, 4, 4, 4, 4, 4, 4,
	3, 3, 3, 3, 3, 3, 3, 3,
	2, 2, 2, 2, 2, 2, 2, 2,
	1, 1, 1, 1, 1, 1, 1, 1,
	0, 0, 0, 0, 0, 0, 0, 0
};
int *row;

/*
the column numbers used by the program
*/

byte_t column_table[64] =
{
	0, 1, 2, 3, 4, 5, 6, 7,
	0, 1, 2, 3, 4, 5, 6, 7,
	0, 1, 2, 3, 4, 5, 6, 7,
	0, 1, 2, 3, 4, 5, 6, 7,
	0, 1, 2, 3, 4, 5, 6, 7,
	0, 1, 2, 3, 4, 5, 6, 7,
	0, 1, 2, 3, 4, 5, 6, 7,
	0, 1, 2, 3, 4, 5, 6, 7
};
int *column;

char nota_colour[4] = {'w', 'b', '.', 'x'};
char *colour_as_string[2] = {"white ", "black "};

enum {PAWN = 0, KNIGHT, BISHOP, ROOK, QUEEN, KING};

char nota_kind[6] = {'P', 'N', 'B', 'R', 'Q', 'K'};

/*
Hash stuff. See for example
T.A. Marsland, ICCA vol. 9, no. 1 (1986) pp. 3-19.
T. Warnock and B. Wendroff, ICCA vol. 11, no 1 (1988), pp. 10-12
*/

#define UPDATE_KEY(X, Y) {(X).lo ^= (Y).lo; (X).hi ^= (Y).hi;}

hash_key_t *hash_key[2][6];
hash_key_t *ep_key;
hash_key_t *castle_key;
hash_key_t *threat_key;

/*
To detect double pawns, free pawns, strong pieces etc. we need
information about the pawn structure. We store the pawn structure
in 8 bytes (one for each column) and if a square on a column
is occupied by a pawn, the corresponding row bit in the byte for 
that column is set. The following tables are initialized at startup
bit_count: gives the number of bits set in a byte.
bit_low: gives the lowest bit set in a byte.
bit_high: gives the highest bit set in a byte.
*/

#define NBITS 8
#define NBITS_SIZE (1 << NBITS)

int *bit_count;
int *bit_low;
int *bit_high;
int white_occup[8], black_occup[8];
int *occup[2] = {white_occup, black_occup};

/*
Some more evaluation function macro's. SCORE_LIMIT stands for infinity.
If an endgame is detected which is won, SCORE_ENDGAME is added to the score
to make sure the program makes a transition to the endgame.
Of course, in some positions this leads to strange play: the program 
sacrifices a pawn instead of promoting it to a queen in order to make the 
transition to the endgame.
*/

#define SCORE_INVAL   32000
#define SCORE_LIMIT   28000
#define SCORE_MATE    24000
#define SCORE_ENDGAME 20000

/*
The evaluation function parameters. They just run from 'bad'
to 'good' and increase by 5 or 10 points. Nevertheless 
the program scored 3 out of 6 in the first round of the 
Dutch Computerchess Championships 1993
*/

int weak_queen_score = -270;
int weak_rook_score = -220;
int weak_bishop_score = -170;
int weak_knight_score = -120;
int isolated_pawn_score = -70;
int king_danger_score = -40;
int double_pawn_score = -30;
int lagged_pawn_score = -20;
int bishop_blocks_pawn_score = -15;
int bishop_not_developed_score = -5;
int knight_not_developed_score = -5;
int weak_pawn_attack_score = 10;
int pin_score = 20;
int rook_on_half_open_file_score = 25;
int knight_strong_score = 30;
int bishop_strong_score = 35;
int rook_on_open_file_score = 40;
int rook_on_7th_rank_score = 50;
int long_castle_score = 60;
int short_castle_score = 70;
int knight_can_move_score = 2;
int bishop_can_move_score = 3;
int rook_can_move_score = 1;

int value[6] = 
{
	PAWN_SCORE, KNIGHT_SCORE, BISHOP_SCORE, ROOK_SCORE, QUEEN_SCORE, KING_SCORE
};

/*
capture strength values (see ICCA journal, july 1993)
*/

int strength[6] = 
{
	9, 7, 7, 5, 1, 9
};

/* 
The king gradient is used in simple endgames. It forces the king to the
side of the board.
*/

byte_t king_alone_grad_table[64] =
{
	-64,-32,-16,-16,-16,-16,-32,-64,
	-32,-16, -8, -8, -8, -8,-16,-32,
	-16, -8, -4, -4, -4, -4, -8,-16,
	-16, -8, -4, -2, -2, -4, -8,-16,
	-16, -8, -4, -2, -2, -4, -8,-16,
	-16, -8, -4, -4, -4, -4, -8,-16,
	-32,-16, -8, -8, -8, -8,-16,-32,
	-64,-32,-16,-16,-16,-16,-32,-64
};

int *king_alone_grad;

/*
The pawn gradient urges pawns to promote.
*/

byte_t free_pawn_grad_table[64] =
{
	 0,  0,  0,  0,  0,  0,  0,  0,
	64, 96, 96, 96, 96, 96, 96, 64,
	32, 48, 48, 48, 48, 48, 48, 32,
	16, 24, 24, 24, 24, 24, 24, 16,
	 8, 12, 12, 12, 12, 12, 12,  8,
	 4,  6,  6,  6,  6,  6,  6,  4,
	 2,  3,  3,  3,  3,  3,  3,  2,
	 0,  0,  0,  0,  0,  0,  0,  0
};

int *white_free_pawn_grad, *black_free_pawn_grad;
int *free_pawn_grad[2];

/*
Visit tables for pawn and knight (only some fields are used in this version)
*/

byte_t pawn_visit_table[64] = 
{
	 0,  0,  0,  0,  0,  0,  0,  0,
	 0,  0,  0,  0,  0,  0,  0,  0,
	 0,  0,  0,  0,  0,  0,  0,  0,
	 0,  0,  0,  0,  0,  0,  0,  0,
	 0,  0,  0,  3,  3,  0,  0,  0,
	 0,  0,  1,  2,  2,  1,  0,  0,
	 0,  0,  0,  0,  0,  0,  0,  0,
	 0,  0,  0,  0,  0,  0,  0,  0
};

int *white_pawn_visit, *black_pawn_visit;
int *pawn_visit[2];

byte_t knight_visit_table[64] = 
{
	 0,  0,  0,  0,  0,  0,  0,  0,
	 0,  0,  0,  0,  0,  0,  0,  0,
	 0,  0,  0,  0,  0,  0,  0,  0,
	 0,  0,  0,  0,  0,  0,  0,  0,
	 0,  0,  0,  0,  0,  0,  0,  0,
	 0,  0,  3,  0,  0,  3,  0,  0,
	 0,  0,  0,  2,  2,  0,  0,  0,
	 0,  0,  0,  0,  0,  0,  0,  0
};

int *white_knight_visit, *black_knight_visit;
int *knight_visit[2];

int *pawn_board;

/*
rooks on 7-th rank are strong
*/

int strong_row[2] = {6, 1};

/*
attack_zero is used to clear the attack array.
*/

int attack_zero[BOARD_MAX] = 
{
	-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,
	-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,
	-1,-1, 0, 0, 0, 0, 0, 0, 0, 0,-1,-1,
	-1,-1, 0, 0, 0, 0, 0, 0, 0, 0,-1,-1,
	-1,-1, 0, 0, 0, 0, 0, 0, 0, 0,-1,-1,
	-1,-1, 0, 0, 0, 0, 0, 0, 0, 0,-1,-1,
	-1,-1, 0, 0, 0, 0, 0, 0, 0, 0,-1,-1,
	-1,-1, 0, 0, 0, 0, 0, 0, 0, 0,-1,-1,
	-1,-1, 0, 0, 0, 0, 0, 0, 0, 0,-1,-1,
	-1,-1, 0, 0, 0, 0, 0, 0, 0, 0,-1,-1,
	-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,
	-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1
};

int *strength_board;

int *attack_board;

/* movegen stuff */

#define MOVE_MAX 16384

/*
We pack the move (from, to and kind) in a long.
*/

#define PACK_MOVE(X, K, A, B)\
	(X) = ((K) | ((B) << 8) | (A));
#define UNPACK_MOVE(X, K, A, B)\
	(K) = ((X) & 0xFFFF0000L);\
	(A) = (int) ((X) & 0x000000FFL);\
	(B) = (int) (((X) >> 8) & 0x000000FFL);
#define ADD_MOVE(P, K, A, B)\
	{\
		PACK_MOVE(*(P), K, A, B)\
		(P)++;\
	}

/*
We generate moves in a smart order: pawn captures first, then
knight, bishop, rook, queen captures. Then castling moves and
finally non-capturing moves. When we search for captures, we store the
captures in a capture stack, and the normal moves in a move stack.
move_top and capt_top are pointers to the current top of the
capture and move stacks.
*/

move_t *move_stack, *capt_stack;
move_t *move_top, *capt_top;

int shift[6] = 
	{FALSE, FALSE, TRUE, TRUE, TRUE, FALSE};

/*
The offsets to generate the moves. We use the forward directions first.
This may give more cutoffs (the moves are generated in 
a more natural order) but I never checked how much.
The end of the list is denoted by offset 0. So you will often see
loops like

board_t *p;
int *d;
for (p = list[icolour][ikind]; p->board_kind != INVALID; p++)
{
	if (p->board_where == INVALID) continue;
	for (d = p->board_dir; *d != 0; d++)
	{
		..
	}
}

*/

int white_pawn_dir[4] =
	{-12,   -13, -11, 0};
int black_pawn_dir[4] =
	{ 12,    11,  13, 0};
int *pawn_dir[2] =
	{white_pawn_dir, black_pawn_dir};

int white_knight_dir[9] =
	{-25, -23, -14, -10,  10, 14,  23,   25,  0};
int black_knight_dir[9] =
	{ 25,  23,  14,  10, -10, -14, -23, -25,  0};
int *knight_dir[2] =
	{white_knight_dir, black_knight_dir};

int white_bishop_dir[5] =
	{-13, -11,  11,  13, 0};
int black_bishop_dir[5] =
	{ 13,  11, -11, -13, 0};
int *bishop_dir[2] =
	{white_bishop_dir, black_bishop_dir};

int white_rook_dir[5] =
	{-12,  -1,   1,  12, 0};
int black_rook_dir[5] =
	{ 12,   1,  -1, -12, 0};
int *rook_dir[2] =
	{white_rook_dir, black_rook_dir};

int white_queen_dir[9] =
	{-13, -12, -11, -1,   1,  11,  12,  13,  0};
int black_queen_dir[9] =
	{ 13,  12,  11,  1,  -1, -11, -12, -13,  0};
int *queen_dir[2] =
	{white_queen_dir, black_queen_dir};

int white_king_dir[9] = 
	{-13, -12, -11,  -1,  1,  11,  12,  13,  0};
int black_king_dir[9] = 
	{ 13,  12,  11,   1, -1, -11, -12, -13,  0};
int *king_dir[2] =
	{white_king_dir, black_king_dir};

int *white_dir[6] =
	{white_pawn_dir, white_knight_dir, white_bishop_dir, white_rook_dir, white_queen_dir, white_king_dir};
int *black_dir[6] =
	{black_pawn_dir, black_knight_dir, black_bishop_dir, black_rook_dir, black_queen_dir, black_king_dir};
int **dir[2] =
	{white_dir, black_dir};
	
int prom_row[2] = {7, 0};

/*
We store the pieces in lists, one list for every piece.
The board is an array of pointers to these lists. Two special
list entries are the empty square and the side square.
*/

#define LIST_MAX 16

board_t *board[BOARD_MAX];
board_t empty = {NONE, INVALID, 0, INVALID, TRUE, FALSE, NULL, FALSE, NULL};
board_t side = {SIDE, INVALID, 0, INVALID, TRUE, FALSE, NULL, FALSE, NULL};

board_t white_pawn_list[LIST_MAX], black_pawn_list[LIST_MAX];
board_t *pawn_list[2] =
	{white_pawn_list, black_pawn_list};

board_t white_knight_list[LIST_MAX], black_knight_list[LIST_MAX];
board_t *knight_list[2] =
	{white_knight_list, black_knight_list};

board_t white_bishop_list[LIST_MAX], black_bishop_list[LIST_MAX];
board_t *bishop_list[2] =
	{white_bishop_list, black_bishop_list};

board_t white_rook_list[LIST_MAX], black_rook_list[LIST_MAX];
board_t *rook_list[2] =
	{white_rook_list, black_rook_list};

board_t white_queen_list[LIST_MAX], black_queen_list[LIST_MAX];
board_t *queen_list[2] =
	{white_queen_list, black_queen_list};

board_t white_king_list[LIST_MAX], black_king_list[LIST_MAX];
board_t *king_list[2] =
	{white_king_list, black_king_list};

board_t *white_list[6] =
	{white_pawn_list, white_knight_list, white_bishop_list, white_rook_list, white_queen_list, white_king_list};
board_t *black_list[6] =
	{black_pawn_list, black_knight_list, black_bishop_list, black_rook_list, black_queen_list, black_king_list};
board_t **list[2] =
	{white_list, black_list};
	
int white_piece_count[6];
int black_piece_count[6];
int	*piece_count[2] = {white_piece_count, black_piece_count};
int tot_count[2];

#define NODE_MAX 1024

/*
The program uses two hash tables: one for the alpha-beta and one for
the static evaluation. The hash table for the quiescence search is only
used for printing the principle variation.
The hash table sizes are powers of two, so we can calculate the index
by (hask_key.lo & (~(~0L << CACHE_BITS))).
Since we then already know the value of the lower CACHE_BITS bits, 
we can use these bits to store other useful information.
*/

#if MACHINE == MAC
#define CACHE_BITS       13
#define SCORE_CACHE_BITS 13
#define QUIES_CACHE_BITS 13
#else
#define CACHE_BITS       16
#define SCORE_CACHE_BITS 16
#define QUIES_CACHE_BITS 13
#endif

#define CACHE_MASK       (~(~0L << CACHE_BITS))
#define SCORE_CACHE_MASK (~(~0L << SCORE_CACHE_BITS))
#define QUIES_CACHE_MASK (~(~0L << QUIES_CACHE_BITS))

#define CACHE_KIND_BITS 11
#define CACHE_KIND_MASK (~(~0L << CACHE_KIND_BITS))

#if CACHE_BITS < CACHE_KIND_BITS
#error CACHE_BITS should not be smaller than CACHE_KIND_BITS
#endif

/*
We use 6 bits for the depth, so the maximum search depth than can be stored 
in the has table is 64.
*/

#define CACHE_DEPTH_BITS 6
#define CACHE_DEPTH_MASK (~(~0UL << CACHE_DEPTH_BITS))

/*
CACHE_INVAL is set if the hash table entry is empty.
CACHE_PV is set if we are in the principle variation. We use this bit
to prevent overwriting PV hash table entries (this
gives analyze_cache a better chance to come up with a PV).
*/

#define CACHE_INVAL BITL(6)
#define CACHE_LOWER BITL(7)
#define CACHE_UPPER BITL(8)
#define CACHE_TRUE  BITL(9)
#define CACHE_PV    BITL(10)

node_t *node_stack, *node;
int inode;

/*
cache_best is the best move.
cache_score is the score.
cache_key is the key
*/

typedef struct
{
	move_t cache_best;
	int cache_score;
	hash_key_t cache_key;
} cache_t;

/*
For the static evaluation hash table we do not have a best move.
*/

typedef struct
{
	int cache_score;
	hash_key_t cache_key;
} score_cache_t;

/*
For the quiescence hash table we do not use the score (it would be
very dangerous to use a hash table in the quiescence 
search, because the quiescence value is dependent on beta.
You can only do it if you also hash beta into the hash key.)
*/

typedef struct
{
	move_t quies_move;
	hash_key_t quies_key;
} quies_cache_t;

/*
we use separate hash tables for white and black
*/

cache_t *cache[2];
score_cache_t *score_cache[2];
quies_cache_t *quies_cache[2];
book_t *book[2] = {NULL, NULL};
long nbook[2] = {0, 0};
char *book_name[2] = {"zwbook.bin", "zbbook.bin"};

int hack_ep;

/*
If null_move is TRUE, a null_move is in effect.
If principle_variation is TRUE, we are in the principle variation.
nthreats is the number of threat extensions in the current line.
*/

int null_move, principle_variation;
int nthreats;

/*
Some counters for statistics
*/

long nodes, ngens, ndos;
long nevals, nhits, nextends;

/*
Variables for time control
*/

clock_t clock_begin, clock_end;
double time_used;

/*
my_printf works just like printf, but optionally echoes to a file
(and flushes the buffer, so that you can still see what went wrong
just before the core dump)
*/

void my_printf(char *format, ...)
{
	va_list arg;
#if MACHINE == MAC
	static FILE *stdecho = NULL;
#endif
	
	va_start(arg, format);
#if OUTPUT > 0
	vfprintf(stdout, format, arg);
	fflush(stdout);
#endif
#if MACHINE == MAC
	if (stdecho == NULL)
		BUG((stdecho = fopen("echo", "w")) == NULL)
	vfprintf(stdecho, format, arg);
	fflush(stdecho);
#endif
	va_end(arg);
}

/*
A 32 bit random number generator. An implementation in C of the algorithm given by
Knuth, the art of computer programming, vol. 2, pp. 26-27. We use e=32, so 
we have to evaluate y(n) = y(n - 24) + y(n - 55) mod 2^32, which is implicitly
done by unsigned arithmetic.
*/

ulong_t ulrand(void)
{
	/*
	random numbers from Mathematica 2.0.
	SeedRandom = 1;
	Table[Random[Integer, {0, 2^32 - 1}]
	*/
	static ulong_t x[55] =
	{
		1410651636UL,
		3012776752UL,
		3497475623UL,
		2892145026UL,
		1571949714UL,
		3253082284UL,
		3489895018UL,
		387949491UL, 
		2597396737UL,
		1981903553UL,
		3160251843UL,
		129444464UL, 
		1851443344UL,
		4156445905UL,
		224604922UL,
		1455067070UL, 
		3953493484UL,
		1460937157UL,
		2528362617UL,
		317430674UL, 
		3229354360UL,
		117491133UL,
		832845075UL,
		1961600170UL, 
		1321557429UL,
		747750121UL,
		545747446UL,
		810476036UL,
		503334515UL, 
		4088144633UL,
		2824216555UL,
		3738252341UL,
		3493754131UL, 
		3672533954UL,
		29494241UL,
		1180928407UL,
		4213624418UL,
		33062851UL, 
		3221315737UL,
		1145213552UL,
		2957984897UL,
		4078668503UL, 
		2262661702UL,
		65478801UL,
		2527208841UL,
		1960622036UL,
		315685891UL, 
		1196037864UL,
		804614524UL,
		1421733266UL,
		2017105031UL, 
		3882325900UL,
		810735053UL,
		384606609UL,
		2393861397UL
	};
	static int init = TRUE;
	static ulong_t y[55];
	static int j, k;
	ulong_t ul;
	
	if (init)
	{
		int i;
		
		init = FALSE;
		for (i = 0; i < 55; i++)
			y[i] = x[i];
		j = 24 - 1;
		k = 55 - 1;
	}
	
	ul = (y[k] += y[j]);
	if (--j < 0) j = 55 - 1;
	if (--k < 0) k = 55 - 1;
	return(ul);
}

void display_move(char *, move_t, char *);
void display_move(char *prefix, move_t m, char *postfix)
{
	long kind;
	int from, to;
	char c;
	
	my_printf("%s", prefix);
	UNPACK_MOVE(m, kind, from, to)
	if (kind == NULL_MOVE)
	{
		my_printf("NUL");
	}
	else if (kind & PROM_CAPT)
	{
		if (kind & PROM_QUEEN)
			c = 'Q';
		else if (kind & PROM_ROOK)
			c = 'R';
		else if (kind & PROM_BISHOP)
			c = 'B';
		else
			c = 'N';
		my_printf("%sx%s%c", GET_NOTA(from), GET_NOTA(to), c);
	}
	else if (kind & PROM_NORM)
	{
		if (kind & PROM_QUEEN)
			c = 'Q';
		else if (kind & PROM_ROOK)
			c = 'R';
		else if (kind & PROM_BISHOP)
			c = 'B';
		else
			c = 'N';
		my_printf("%s-%s%c", GET_NOTA(from), GET_NOTA(to), c);
	}
	else if (kind == EP)
		my_printf("%sx%s e.p.", GET_NOTA(from), GET_NOTA(to));
	else if (kind == SHORT)
		my_printf("o-o");
	else if (kind == LONG)
		my_printf("o-o-o");
	else if (kind == CAPT)
		my_printf("%sx%s", GET_NOTA(from), GET_NOTA(to));
	else
		my_printf("%s-%s", GET_NOTA(from), GET_NOTA(to));
	my_printf("%s", postfix);
}

void move_to_string(move_t, char *);
void move_to_string(move_t m, char *s)
{
	long kind;
	int from, to;
	char c;
	
	UNPACK_MOVE(m, kind, from, to)
	if (kind == NULL_MOVE)
	{
		sprintf("%s", "NUL");
	}
	else if (kind & PROM_CAPT)
	{
		if (kind & PROM_QUEEN)
			c = 'Q';
		else if (kind & PROM_ROOK)
			c = 'R';
		else if (kind & PROM_BISHOP)
			c = 'B';
		else
			c = 'N';
		sprintf(s, "%s%s%c", GET_NOTA(from), GET_NOTA(to), c);
	}
	else if (kind & PROM_NORM)
	{
		if (kind & PROM_QUEEN)
			c = 'Q';
		else if (kind & PROM_ROOK)
			c = 'R';
		else if (kind & PROM_BISHOP)
			c = 'B';
		else
			c = 'N';
		sprintf(s, "%s%s%c", GET_NOTA(from), GET_NOTA(to), c);
	}
	else if (kind == EP)
		sprintf(s, "%s%s", GET_NOTA(from), GET_NOTA(to));
	else if (kind == SHORT)
		sprintf(s, "o-o");
	else if (kind == LONG)
		sprintf(s, "o-o-o");
	else if (kind == CAPT)
		sprintf(s, "%s%s", GET_NOTA(from), GET_NOTA(to));
	else
		sprintf(s, "%s%s", GET_NOTA(from), GET_NOTA(to));
}

void display_board(void)
{
	int i, j, k;
	
	k = 0;
	for (j = 0; j < 8; j++)
	{
		for (i = 0; i < 8; i++)
		{
			if (board[map[k]] == &empty)
				my_printf("..");
			else
				my_printf("%c%c",
					nota_colour[board[map[k]]->board_colour],
					nota_kind[board[map[k]]->board_kind]);
			k++;
		}
		my_printf("\n");
	}
}

/*
write_board writes the board to a file.
*/

void write_board(FILE *f)
{
	int i, j, k;
	
	k = 0;
	for (j = 0; j < 8; j++)
	{
		for (i = 0; i < 8; i++)
		{
			if (board[map[k]] == &empty)
				fprintf(f, "..");
			else
				fprintf(f, "%c%c",
					nota_colour[board[map[k]]->board_colour],
					nota_kind[board[map[k]]->board_kind]);
			k++;
		}
		fprintf(f, "\n");
	}
}

/*
display_list displays the lists of pieces. Captured pieces are
indicated by ??
*/

void display_list(void)
{
	int icolour;
	int ikind;
	board_t *p;

	for (icolour = WHITE; icolour <= BLACK; icolour++)
	{
		my_printf("%s%d", colour_as_string[icolour], tot_count[icolour]);
		for (ikind = PAWN; ikind <= KING; ikind++)
		{
			my_printf(" %d%c", piece_count[icolour][ikind], nota_kind[ikind]);
			for (p = list[icolour][ikind]; p->board_kind != INVALID; p++)
			{
				if (p->board_where == INVALID)
					my_printf(" ??");
				else
					my_printf(" %s", GET_NOTA(p->board_where));
			}
		}
		my_printf("\n");
	}
}

/*
display_occup displays the occupation numbers for the columns.
*/

void display_occup(void)
{
	int icolour;
	int icolumn;
	
	for (icolour = BLACK; icolour >= WHITE; icolour--)
	{
		my_printf("%s", colour_as_string[icolour]);
		for (icolumn = 0; icolumn < 8; icolumn++)
			my_printf(" %d", bit_count[occup[icolour][icolumn]]);
		my_printf("\n");
	}
}

/*
display_attack displays the attack table.
*/

void display_attack(void)
{
	int i, j, k;
	
	k = 0;
	for (j = 0; j < 8; j++)
	{
		for (i = 0; i < 8; i++)
		{
			my_printf(" %2d", attack_board[map[k]]);
			k++;
		}
		my_printf("\n");
	}
}

/*
display_strength displays the capture strength table.
*/

void display_strength(void);
void display_strength(void)
{
	int i, j, k;
	
	k = 0;
	for (j = 0; j < 8; j++)
	{
		for (i = 0; i < 8; i++)
		{
			my_printf(" %2d", strength_board[map[k]]);
			k++;
		}
		my_printf("\n");
	}
}

/*
return_mat_score calculates the current material value
*/

int return_mat_score(int);
int return_mat_score(int my_colour)
{
	int sum;
	int ikind;
	board_t *p;
	
	sum = 0;
	for (ikind = PAWN; ikind <= QUEEN; ikind++)
	{
		for (p = list[my_colour][ikind]; p->board_kind != INVALID; p++)
			if (p->board_where != INVALID) sum += p->board_value;
	}
	return(sum);
}

/*
return_cur_key returns the current hash key.
*/

hash_key_t return_cur_key(void);
hash_key_t return_cur_key(void)
{
	int icolour, ikind;
	hash_key_t key;
	board_t *p;
	
	key.lo = key.hi = 0UL;
	for (icolour = WHITE; icolour <= BLACK; icolour++)
	{
		for (ikind = PAWN; ikind <= KING; ikind++)
		{
			for (p = list[icolour][ikind]; p->board_kind != INVALID; p++)
			{
				if (p->board_where != INVALID)
					UPDATE_KEY(key, p->board_key[p->board_where])
			}
		}
	}
	return(key);
}

/*
update_key updates the hash key with ep rights, castling rights and
the number of threat extensions. This makes the position unique.
*/

void update_key(int your_colour, hash_key_t *key)
{
	int j;
	
	if (hack_ep != INVALID)
	{
		j = hack_ep - 1;
		if ((board[j]->board_colour == your_colour) and (board[j]->board_kind == PAWN))
			UPDATE_KEY(*key, ep_key[j])
		j = hack_ep + 1;
		if ((board[j]->board_colour == your_colour) and (board[j]->board_kind == PAWN))
			UPDATE_KEY(*key, ep_key[j])
	}
	if (!board[E1]->board_moved)
	{
		if (!board[H1]->board_moved)
			UPDATE_KEY(*key, castle_key[H1])
		if (!board[A1]->board_moved)
			UPDATE_KEY(*key, castle_key[A1])
	}
	if (!board[E8]->board_moved)
	{
		if (!board[H8]->board_moved)
			UPDATE_KEY(*key, castle_key[H8])
		if (!board[A1]->board_moved)
			UPDATE_KEY(*key, castle_key[A8])
	}
	UPDATE_KEY(*key, threat_key[nthreats])
}

/*
check_lists checks whether the lists of pieces are still correct.
*/

int check_list(void);
int check_list(void)
{
	int i;
	int icolour;
	board_t *p;
	int n, m;
	int ikind;
	
	for (i = 0; i < 64; i++)
		if ((board[map[i]] != &empty) and (board[map[i]]->board_where != map[i])) return(1);
	for (icolour = WHITE; icolour <= BLACK; icolour++)
	{
		n = 0;
		for (ikind = PAWN; ikind <= KING; ikind++)
		{
			m = 0;
			for (p = list[icolour][ikind]; p->board_kind != INVALID; p++)
			{
				if (p->board_where != INVALID)
				{
					if (board[p->board_where] != p) return(2);
					m++;
				}
			}
			if (m != piece_count[icolour][ikind]) return(3);
			n += m;
		}
		if (n != tot_count[icolour]) return(4);
	}
	return(0);
}

/*
check_occup checks the occupation numbers.
*/

int check_occup(void);
int check_occup(void)
{
	int temp[8];
	int icolour;
	int icolumn;
	board_t *p;
	
	for (icolour = WHITE; icolour <= BLACK; icolour++)
	{
		for (icolumn = 0; icolumn < 8; icolumn++)
			temp[icolumn] = 0;
		for (p = pawn_list[icolour]; p->board_kind != INVALID; p++)
			if (p->board_where != INVALID) temp[column[p->board_where]] |= (1 << row[p->board_where]);
		for (icolumn = 0; icolumn < 8; icolumn++)
			if (temp[icolumn] != occup[icolour][icolumn]) return(1);
	}
	return(0);
}

/*
compact_lists removes captured pieces from the lists. 
Is called before we start thinking about a position.
*/

void compact_lists(void)
{
	int icolour, ikind;
	board_t *p;
	int err;
	int n, i;
	board_t temp[LIST_MAX];
	
	for (icolour = WHITE; icolour <= BLACK; icolour++)
	{
		for (ikind = PAWN; ikind <= QUEEN; ikind++)
		{
			/*
			copy all pieces to temp
			*/
			
			n = 0;
			for (p = list[icolour][ikind]; p->board_kind != INVALID; p++)
			{
				if (p->board_where != INVALID)
				{
					BUG(n >= LIST_MAX)
					temp[n] = *p;
					n++;
				}
			}
			p = list[icolour][ikind];
			/*
			copy back.
			*/
			
			for (i = 0; i < n; i++)
			{
				*p = temp[i];
				board[p->board_where] = p;
				p++;
			}
			for ( ; i < LIST_MAX; i++)
			{
				p->board_kind = p->board_where = INVALID;
				p++;
			}
		}
	}
	if ((err = check_list()) > 0)
	{
		my_printf("err=%d\n", err);
		display_board();
		display_list();
		display_occup();
		FATAL_ERROR("list error")
	}
	if ((err = check_occup()) > 0)
	{
		my_printf("err=%d\n", err);
		display_board();
		display_list();
		display_occup();
		FATAL_ERROR("occup error")
	}
}

/*
read_board reads a board from file and initializes lists and nodes.
returns FALSE if no board was found.
*/

int read_board(FILE *f)
{
	int i;
	int temp_colour[64], temp_kind[64];
	char temp[LINE_MAX];
	int icolour, ikind, icolumn;
	board_t *p;
	int n;
	int err;
	int start_colour[64] = 
	{
		BLACK, BLACK, BLACK, BLACK, BLACK, BLACK, BLACK, BLACK,
		BLACK, BLACK, BLACK, BLACK, BLACK, BLACK, BLACK, BLACK,
		NONE , NONE , NONE , NONE , NONE , NONE , NONE , NONE ,
		NONE , NONE , NONE , NONE , NONE , NONE , NONE , NONE ,
		NONE , NONE , NONE , NONE , NONE , NONE , NONE , NONE ,
		NONE , NONE , NONE , NONE , NONE , NONE , NONE , NONE ,
		WHITE, WHITE, WHITE, WHITE, WHITE, WHITE, WHITE, WHITE,
		WHITE, WHITE, WHITE, WHITE, WHITE, WHITE, WHITE, WHITE,
	};
	int start_kind[64] = 
	{
		ROOK   , KNIGHT , BISHOP , QUEEN  , KING   , BISHOP , KNIGHT , ROOK   ,
		PAWN   , PAWN   , PAWN   , PAWN   , PAWN   , PAWN   , PAWN   , PAWN   ,
		INVALID, INVALID, INVALID, INVALID, INVALID, INVALID, INVALID, INVALID,
		INVALID, INVALID, INVALID, INVALID, INVALID, INVALID, INVALID, INVALID,
		INVALID, INVALID, INVALID, INVALID, INVALID, INVALID, INVALID, INVALID,
		INVALID, INVALID, INVALID, INVALID, INVALID, INVALID, INVALID, INVALID,
		PAWN   , PAWN   , PAWN   , PAWN   , PAWN   , PAWN   , PAWN   , PAWN   ,
		ROOK   , KNIGHT , BISHOP , QUEEN  , KING   , BISHOP , KNIGHT , ROOK
	};

	hack_ep = INVALID;
	for (i = 0; i < 64; i++)
	{
		if (fscanf(f, "%2s", temp) != 1) return(FALSE);
		if (temp[0] == 'q') return(FALSE);
		if (temp[0] == '.')
		{
			temp_colour[i] = NONE;
			temp_kind[i] = INVALID;
		}
		else
		{
			if (temp[0] == 'w')
				temp_colour[i] = WHITE;
			else if (temp[0] == 'b')
				temp_colour[i] = BLACK;
			else
				FATAL_ERROR("INVALID_COLOUR")
			if (temp[1] == 'K')
				temp_kind[i] = KING;
			else if (temp[1] == 'P')
				temp_kind[i] = PAWN;
			else if (temp[1] == 'N')
				temp_kind[i] = KNIGHT;
			else if (temp[1] == 'B')
				temp_kind[i] = BISHOP;
			else if (temp[1] == 'R')
				temp_kind[i] = ROOK;
			else if (temp[1] == 'Q')
				temp_kind[i] = QUEEN;
			else
				FATAL_ERROR("INVALID_KIND")
		}
	}
	for (i = 0; i < BOARD_MAX; i++)
		board[i] = &side;
	for (i = 0; i < 64; i++)
		board[map[i]] = &empty;
	
	for (icolour = WHITE; icolour <= BLACK; icolour++)
	{
		tot_count[icolour] = 0;
		for (icolumn = 0; icolumn < 8; icolumn++)
			occup[icolour][icolumn] = 0;
		for (ikind = PAWN; ikind <= KING; ikind++)
		{
			piece_count[icolour][ikind] = 0;
			p = list[icolour][ikind];
			n = 0;
			for (i = 0; i < 64; i++)
			{
				if ((temp_colour[i] == icolour) and (temp_kind[i] == ikind))
				{
					tot_count[icolour]++;
					piece_count[icolour][ikind]++;
					p->board_colour = icolour;
					p->board_kind = ikind;
					p->board_value = value[ikind];
					p->board_where = map[i];
					if ((temp_colour[i] == start_colour[i]) and
					    (temp_kind[i] == start_kind[i]))
						p->board_moved = FALSE;
					else
						p->board_moved = TRUE;
					p->board_captured = FALSE;
					p->board_key = hash_key[icolour][ikind];
					p->board_shift = shift[ikind];
					p->board_dir = dir[icolour][ikind];
					p->board_count = piece_count[icolour] + ikind;
					p->board_strength = strength[ikind];
					if (ikind == PAWN)
						occup[icolour][column[p->board_where]] |= (1 << row[p->board_where]);
					board[map[i]] = p;
					p++;
					n++;
				}
			}
			for ( ; n < LIST_MAX; n++)
			{
				p->board_colour = icolour;
				p->board_kind = INVALID;
				p->board_value = value[ikind];
				p->board_where = INVALID;
				p->board_moved = TRUE;
				p->board_captured = FALSE;
				p->board_key = hash_key[icolour][ikind];
				p->board_shift = shift[ikind];
				p->board_dir = dir[icolour][ikind];
				p->board_count = piece_count[icolour] + ikind;
				p->board_strength = strength[ikind];
				p++;
			}
		}
	}
	/*
	set node
	*/
	
	node = node_stack;
	node->move_base = move_stack;
	node->capt_base = capt_stack;
	node->node_key = return_cur_key();
	node->node_total_key = node->node_key;
	node->node_mat[WHITE] = return_mat_score(WHITE);
	node->node_mat[BLACK] = return_mat_score(BLACK);
	node->node_pos[WHITE] = node->node_pos[BLACK] = 0;
	node->node_reverse = FALSE;
	node->node_hack_ep = INVALID;
	null_move = FALSE;
	nthreats = 0;
	
	if ((err = check_list()) > 0)
	{
		my_printf("err=%d\n", err);
		display_board();
		display_list();
		display_occup();
		FATAL_ERROR("list error")
	}
	if ((err = check_occup()) > 0)
	{
		my_printf("err=%d\n", err);
		display_board();
		display_list();
		display_occup();
		FATAL_ERROR("occup error")
	}
	return(TRUE);
}

/*
attacked returns TRUE if square i is attacked by your_colour.
*/

int attacked(int i, int your_colour)
{
	int *d;
	board_t *p;
	int j;
	
	p = board[i - pawn_dir[your_colour][1]];
	if ((p->board_colour == your_colour) and (p->board_kind == PAWN)) return(TRUE);
	p = board[i - pawn_dir[your_colour][2]];
	if ((p->board_colour == your_colour) and (p->board_kind == PAWN)) return(TRUE);
	for (d = knight_dir[your_colour]; *d != 0; d++)
	{
		p = board[i - *d];
		if ((p->board_colour == your_colour) and (p->board_kind == KNIGHT)) return(TRUE);
	}
	for (d = bishop_dir[your_colour]; *d != 0; d++)
	{
		j = i - *d;
		p = board[j];
		if ((p->board_colour == your_colour) and (p->board_kind == KING)) return(TRUE);
		while(board[j] == &empty) j -= *d;
		p = board[j];
		if ((p->board_colour == your_colour) and
		    ((p->board_kind == BISHOP) or (p->board_kind == QUEEN))) return(TRUE);
	}
	for (d = rook_dir[your_colour]; *d != 0; d++)
	{
		j = i - *d;
		p = board[j];
		if ((p->board_colour == your_colour) and (p->board_kind == KING)) return(TRUE);
		while(board[j] == &empty) j -= *d;
		p = board[j];
		if ((p->board_colour == your_colour) and
		    ((p->board_kind == ROOK) or (p->board_kind == QUEEN))) return(TRUE);
	}
	return(FALSE);
}

int i_am_check(int my_colour, int your_colour)
{
	return(attacked(king_list[my_colour]->board_where, your_colour));
}

#define USE_STRENGTH

#ifdef USE_STRENGTH
#define ADD_STRENGTH(P, A, B)\
	*(P)++ = strength_board[B];
#else
#define ADD_STRENGTH(P, A, B)
#endif
	
int *capt_strength_table;
int *capt_strength;
int *move_strength_table;
int *move_strength;

/*
gen_moves generates the moves for the piece on square p.
*/

int gen_moves(int, int, board_t *);
int gen_moves(int my_colour, int your_colour, board_t *p)
{
	int j;
	int *d;
	int irow;
	
	if (p->board_kind == PAWN)
	{
		/*
		capturing moves for pawn
		*/
		j = p->board_where + p->board_dir[1];
		irow = row[j];
		if (board[j]->board_colour == your_colour)
		{
			if (irow == prom_row[my_colour])
			{
				ADD_MOVE(capt_top, (PROM_CAPT | PROM_QUEEN ), p->board_where, j)
				ADD_STRENGTH(capt_strength, p->board_where, j)
				ADD_MOVE(capt_top, (PROM_CAPT | PROM_ROOK  ), p->board_where, j)
				ADD_STRENGTH(capt_strength, p->board_where, j)
				ADD_MOVE(capt_top, (PROM_CAPT | PROM_BISHOP), p->board_where, j)
				ADD_STRENGTH(capt_strength, p->board_where, j)
				ADD_MOVE(capt_top, (PROM_CAPT | PROM_KNIGHT), p->board_where, j)
				ADD_STRENGTH(capt_strength, p->board_where, j)
			}
			else
			{
				ADD_MOVE(capt_top, CAPT, p->board_where, j)
				ADD_STRENGTH(capt_strength, p->board_where, j)
			}
		}
		j = p->board_where + p->board_dir[2];
		irow = row[j];
		if (board[j]->board_colour == your_colour)
		{
			if (irow == prom_row[my_colour])
			{
				ADD_MOVE(capt_top, (PROM_CAPT | PROM_QUEEN ), p->board_where, j)
				ADD_STRENGTH(capt_strength, p->board_where, j)
				ADD_MOVE(capt_top, (PROM_CAPT | PROM_ROOK  ), p->board_where, j)
				ADD_STRENGTH(capt_strength, p->board_where, j)
				ADD_MOVE(capt_top, (PROM_CAPT | PROM_BISHOP), p->board_where, j)
				ADD_STRENGTH(capt_strength, p->board_where, j)
				ADD_MOVE(capt_top, (PROM_CAPT | PROM_KNIGHT), p->board_where, j)
				ADD_STRENGTH(capt_strength, p->board_where, j)
			}
			else
			{
				ADD_MOVE(capt_top, CAPT, p->board_where, j)
				ADD_STRENGTH(capt_strength, p->board_where, j)
			}
		}
		/*
		non-capturing moves for pawn
		*/
		j = p->board_where + p->board_dir[0];
		if (board[j] == &empty)
		{
			irow = row[j];
			if (irow == prom_row[my_colour])
			{
				ADD_MOVE(move_top, (PROM_NORM | PROM_QUEEN ), p->board_where, j)
				ADD_STRENGTH(move_strength, p->board_where, j)
				ADD_MOVE(move_top, (PROM_NORM | PROM_ROOK  ), p->board_where, j)
				ADD_STRENGTH(move_strength, p->board_where, j)
				ADD_MOVE(move_top, (PROM_NORM | PROM_BISHOP), p->board_where, j)
				ADD_STRENGTH(move_strength, p->board_where, j)
				ADD_MOVE(move_top, (PROM_NORM | PROM_KNIGHT), p->board_where, j)
				ADD_STRENGTH(move_strength, p->board_where, j)
			}
			else
			{
				ADD_MOVE(move_top, NORM, p->board_where, j)
				ADD_STRENGTH(move_strength, p->board_where, j)
			}
			/*
			double pawn move
			*/
			if (!p->board_moved)
			{
				j += p->board_dir[0];
				if (board[j] == &empty)
				{
					ADD_MOVE(move_top, DOUBLE, p->board_where, j)
					ADD_STRENGTH(move_strength, p->board_where, j)
				}
			}
		}
	}
	else
	{
		/*
		other pieces
		*/
		for (d = p->board_dir; *d != 0; d++)
		{
			j = p->board_where;
			while(TRUE)
			{
				/*
				try next square
				*/
				j += *d;
				if (board[j] == &empty)
				{
					ADD_MOVE(move_top, NORM, p->board_where, j)
					ADD_STRENGTH(move_strength, p->board_where, j)
					if (!p->board_shift) break;
				}
				else
				{
					/*
					may be capture
					*/
					if (board[j]->board_colour == your_colour)
					{
						ADD_MOVE(capt_top, CAPT, p->board_where, j)
						ADD_STRENGTH(capt_strength, p->board_where, j)
					}
					break;
				}
			}
		}
	}
	return(capt_top - node->capt_base);
}

/*
gen_special_moves generates en passant and short or long castle
*/

int gen_special_moves(int, int);
int gen_special_moves(int my_colour, int your_colour)
{
	int j;
	
	if (hack_ep != INVALID)
	{
		/*
		pawn to the left?
		*/
		j = hack_ep - 1;
		if ((board[j]->board_colour == my_colour) and (board[j]->board_kind == PAWN))
		{
			ADD_MOVE(capt_top, EP, j, hack_ep + pawn_dir[my_colour][0])
		}
		/* 
		pawn to the right?
		*/
		j = hack_ep + 1;
		if ((board[j]->board_colour == my_colour) and (board[j]->board_kind == PAWN))
		{
			ADD_MOVE(capt_top, EP, j, hack_ep + pawn_dir[my_colour][0])
		}
	}
	if (my_colour == WHITE)
	{
		/*
		empty square has board_moved set to TRUE. Dirty trick eh?
		*/
		
		if (!board[E1]->board_moved)
		{
			if (!board[H1]->board_moved)
			{
				if ((board[F1] == &empty) and (board[G1] == &empty) and
				    !attacked(E1, your_colour) and !attacked(F1, your_colour) and !attacked(G1, your_colour))
				{
					ADD_MOVE(capt_top, SHORT, 0, 0)
				}
			}
			if (!board[A1]->board_moved)
			{
				if ((board[D1] == &empty) and (board[C1] == &empty) and
				    (board[B1] == &empty) and
				    !attacked(E1, your_colour) and !attacked(D1, your_colour) and !attacked(C1, your_colour))
				{
					ADD_MOVE(capt_top, LONG, 0, 0)
				}
			}
		}
	}
	else
	{
		if (!board[E8]->board_moved)
		{
			if (!board[H8]->board_moved)
			{
				if ((board[F8] == &empty) and (board[G8] == &empty) and
				    !attacked(E8, your_colour) and !attacked(F8, your_colour) and !attacked(G8, your_colour))
				{  
					ADD_MOVE(capt_top, SHORT, 0, 0)
				}
			}
			if (!board[A8]->board_moved)
			{
				if ((board[D8] == &empty) and (board[C8] == &empty) and
				    (board[B8] == &empty) and
				    !attacked(E8, your_colour) and !attacked(D8, your_colour) and !attacked(C8, your_colour))
				{
					ADD_MOVE(capt_top, LONG, 0, 0)
				}
			}
		}
	}
	return(capt_top - node->capt_base);
}

/*
gen_capts generates all captures to a square. Same as attacked routine,
except ADD_MOVE is added.
*/

int gen_capts(int, int);
int gen_capts(int i, int your_colour)
{
	int *d;
	board_t *p;
	int j;
	
	j = i - pawn_dir[your_colour][1];
	p = board[j];
	if ((p->board_colour == your_colour) and (p->board_kind == PAWN))
	{
		if (row[i] == prom_row[your_colour])
		{
			ADD_MOVE(capt_top, (PROM_CAPT | PROM_QUEEN ), j, i)
			ADD_MOVE(capt_top, (PROM_CAPT | PROM_ROOK  ), j, i)
			ADD_MOVE(capt_top, (PROM_CAPT | PROM_BISHOP), j, i)
			ADD_MOVE(capt_top, (PROM_CAPT | PROM_KNIGHT), j, i)
		}
		else
			ADD_MOVE(capt_top, CAPT, j, i)
	}
	j = i - pawn_dir[your_colour][2];
	p = board[j];
	if ((p->board_colour == your_colour) and (p->board_kind == PAWN))
	{
		if (row[i] == prom_row[your_colour])
		{
			ADD_MOVE(capt_top, (PROM_CAPT | PROM_QUEEN ), j, i)
			ADD_MOVE(capt_top, (PROM_CAPT | PROM_ROOK  ), j, i)
			ADD_MOVE(capt_top, (PROM_CAPT | PROM_BISHOP), j, i)
			ADD_MOVE(capt_top, (PROM_CAPT | PROM_KNIGHT), j, i)
		}
		else
			ADD_MOVE(capt_top, CAPT, j, i)
	}
	for (d = knight_dir[your_colour]; *d != 0; d++)
	{
		j = i - *d;
		p = board[j];
		if ((p->board_colour == your_colour) and (p->board_kind == KNIGHT)) ADD_MOVE(capt_top, CAPT, j, i)
	}
	for (d = bishop_dir[your_colour]; *d != 0; d++)
	{
		j = i - *d;
		p = board[j];
		if ((p->board_colour == your_colour) and (p->board_kind == KING)) ADD_MOVE(capt_top, CAPT, j, i)
		while(board[j] == &empty) j -= *d;
		p = board[j];
		if ((p->board_colour == your_colour) and
		    ((p->board_kind == BISHOP) or (p->board_kind == QUEEN))) ADD_MOVE(capt_top, CAPT, j, i)
	}
	for (d = rook_dir[your_colour]; *d != 0; d++)
	{
		j = i - *d;
		p = board[j];
		if ((p->board_colour == your_colour) and (p->board_kind == KING)) ADD_MOVE(capt_top, CAPT, j, i)
		while(board[j] == &empty) j -= *d;
		p = board[j];
		if ((p->board_colour == your_colour) and
		    ((p->board_kind == ROOK) or (p->board_kind == QUEEN))) ADD_MOVE(capt_top, CAPT, j, i)
	}
	return(capt_top - node->capt_base);
}

/*
Chess do_move routines always seem to get a bit messy when you add features
like occupation numbers, piece lists and piece counts..
*/

void do_move(int my_colour, int your_colour, move_t m)
{
/*
DO_CAPT removes a piece on SQUARE B. 
*/

#define DO_CAPT(B)\
	{\
		p = board[B];\
		BUG(p->board_colour != your_colour)\
		BUG(p->board_kind == KING)\
		node->capt = p;\
		p->board_where = INVALID;\
		if (p->board_kind == PAWN)\
			occup[your_colour][column[B]] &= ~(1 << row[B]);\
		UPDATE_KEY(node->node_key, p->board_key[B])\
		node->node_mat[your_colour] -= p->board_value;\
		tot_count[your_colour]--;\
		(*p->board_count)--;\
	}

/*
DO_MOVE moves a piece from square A to square B.
*/

#define DO_MOVE(A, B)\
	{\
		p = board[A];\
		UPDATE_KEY(node->node_key, p->board_key[A])\
		UPDATE_KEY(node->node_key, p->board_key[B])\
		p->board_where = B;\
		node->node_moved = p->board_moved;\
		node->node_captured = p->board_captured;\
		p->board_moved = TRUE;\
		board[A] = &empty;\
		board[B] = p;\
	}

	board_t *p;
	int ws, wl, bs, bl;
	long kind;
	int from, to;
	int ep;
#if DEBUG > 0
	int err;
	hash_key_t temp_key;
#endif
	
	/*
	It is difficult to check whether white or black has castled in an end node.
	So before the move we check the castling rights, so that we can give
	a bonus for castling (or a penalty for losing castling rights).
	*/
	
	ws = wl = bs = bl = FALSE;
	if (!board[E1]->board_moved)
	{
		if (!board[H1]->board_moved) ws = TRUE;
		if (!board[A1]->board_moved) wl = TRUE;
	}
	if (!board[E8]->board_moved)
	{
		if (!board[H8]->board_moved) bs = TRUE;
		if (!board[A8]->board_moved) bl = TRUE;
	}
	
	/*
	Save hack_ep for undo move
	*/
	
	node->node_move = m;
	node->node_hack_ep = hack_ep;
	hack_ep = INVALID;
	
	/*
	Copy incrementally updated information to next node.
	*/
	node[1].node_key = node->node_key;
	node[1].node_mat[WHITE] = node->node_mat[WHITE];
	node[1].node_mat[BLACK] = node->node_mat[BLACK];
	node[1].node_pos[WHITE] = node->node_pos[WHITE];
	node[1].node_pos[BLACK] = node->node_pos[BLACK];
	node++;
	inode++;
	
	node->node_reverse = FALSE;
	
	UNPACK_MOVE(m, kind, from, to)
	if (kind == NULL_MOVE) goto label_null;
	
	/*
	Another cheap check.
	*/
	
	if ((kind != SHORT) and (kind != LONG))
	{
		BUG(column[from] == INVALID)
		BUG(column[to] == INVALID)
	}
	if (kind == EP)
	{
		ep = to - pawn_dir[my_colour][0];
		DO_CAPT(ep)
		board[ep] = &empty;
		DO_MOVE(from, to)
		occup[my_colour][column[from]] &= ~(1 << row[from]);
		occup[my_colour][column[to]] |= (1 << row[to]);
	}
	else if (kind & PROM_CAPT)
	{
		DO_CAPT(to)
		
		/*
		Remove pawn
		*/
		
		occup[my_colour][column[from]] &= ~(1 << row[from]);
		p = board[from];
		node->prom = p;
		board[from] = &empty;
		p->board_where = INVALID;
		(*p->board_count)--;
		UPDATE_KEY(node->node_key, p->board_key[from])
		node->node_mat[my_colour] -= p->board_value;
		
		/*
		Add piece. Search end of list first.
		*/
		
		if (kind & PROM_QUEEN)
		{
			for (p = queen_list[my_colour]; p->board_kind == QUEEN; p++);
			p->board_kind = QUEEN;
		}
		else if (kind & PROM_ROOK)
		{
			for (p = rook_list[my_colour]; p->board_kind == ROOK; p++);
			p->board_kind = ROOK;
		}
		else if (kind & PROM_BISHOP)
		{
			for (p = bishop_list[my_colour]; p->board_kind == BISHOP; p++);
			p->board_kind = BISHOP;
		}
		else
		{
			for (p = knight_list[my_colour]; p->board_kind == KNIGHT; p++);
			p->board_kind = KNIGHT;
		}
		p->board_where = to;
		(*p->board_count)++;
		board[to] = p;
		UPDATE_KEY(node->node_key, p->board_key[to])
		node->node_mat[my_colour] += p->board_value;
	}
	else if (kind == SHORT)
	{
		if (my_colour == WHITE)
		{
			DO_MOVE(E1, G1)
			DO_MOVE(H1, F1)
		}
		else
		{
			DO_MOVE(E8, G8)
			DO_MOVE(H8, F8)
		}
	}
	else if (kind == LONG)
	{
		if (my_colour == WHITE)
		{
			DO_MOVE(E1, C1)
			DO_MOVE(A1, D1)
		}
		else
		{
			DO_MOVE(E8, C8)
			DO_MOVE(A8, D8)
		}
	}
	else if (kind & PROM_NORM)
	{
		/*
		Remove pawn
		*/

		occup[my_colour][column[from]] &= ~(1 << row[from]);
		p = board[from];
		node->prom = p;
		board[from] = &empty;
		p->board_where = INVALID;
		(*p->board_count)--;
		UPDATE_KEY(node->node_key, p->board_key[from])
		node->node_mat[my_colour] -= p->board_value;
		
		/*
		Add piece
		*/
		
		if (kind & PROM_QUEEN)
		{
			for (p = queen_list[my_colour]; p->board_kind == QUEEN; p++);
			p->board_kind = QUEEN;
		}
		else if (kind & PROM_ROOK)
		{
			for (p = rook_list[my_colour]; p->board_kind == ROOK; p++);
			p->board_kind = ROOK;
		}
		else if (kind & PROM_BISHOP)
		{
			for (p = bishop_list[my_colour]; p->board_kind == BISHOP; p++);
			p->board_kind = BISHOP;
		}
		else
		{
			for (p = knight_list[my_colour]; p->board_kind == KNIGHT; p++);
			p->board_kind = KNIGHT;
		}
		p->board_where = to;
		(*p->board_count)++;
		board[to] = p;
		UPDATE_KEY(node->node_key, p->board_key[to])
		node->node_mat[my_colour] += p->board_value;
	}
	else
	{
		if (kind == CAPT)
			DO_CAPT(to)
		DO_MOVE(from, to)
		if (p->board_kind == PAWN)
		{
			occup[my_colour][column[from]] &= ~(1 << row[from]);
			occup[my_colour][column[to]] |= (1 << row[to]);
		}
		if (kind == CAPT)
			p->board_captured = TRUE;
		else if (kind == DOUBLE)
			hack_ep = to;
		else
			node->node_reverse = TRUE;
	}
	
	/*
	Give a bonus for castling or a penalty for losing castling rights.
	*/
	
	if (kind == SHORT)
		node->node_pos[my_colour] += short_castle_score;
	else if (kind == LONG)
		node->node_pos[my_colour] += long_castle_score;
	else
	{
		if (ws and (board[E1]->board_moved or board[H1]->board_moved))
			node->node_pos[WHITE] -= short_castle_score;
		if (wl and (board[E1]->board_moved or board[A1]->board_moved))
			node->node_pos[WHITE] -= long_castle_score;
		if (bs and (board[E8]->board_moved or board[H8]->board_moved))
			node->node_pos[BLACK] -= short_castle_score;
		if (bl and (board[E8]->board_moved or board[A8]->board_moved))
			node->node_pos[BLACK] -= long_castle_score;
	}
	
	label_null:
	node->node_total_key = node->node_key;
	update_key(your_colour, &(node->node_total_key));
#if DEBUG > 0
	if ((err = check_list()) > 0)
	{
		my_printf("%s err=%d\n", colour_as_string[my_colour], err);
		display_move("", m, "\n");
		display_board();
		display_list();
		display_occup();
		FATAL_ERROR("do list error")
	}
	temp_key = return_cur_key();
	if ((temp_key.lo != node->node_key.lo) or
	    (temp_key.hi != node->node_key.hi))
	{
		my_printf("%s\n", colour_as_string[my_colour]);
		display_move("", m, "\n");
		display_board();
		display_list();
		display_occup();
		FATAL_ERROR("do key error")
	}
	if ((err = check_occup()) > 0)
	{
		my_printf("%s err=%d\n", colour_as_string[my_colour], err);
		display_move("", m, "\n");
		display_board();
		display_list();
		display_occup();
		FATAL_ERROR("do occup error")
	}
	if (node->node_mat[WHITE] != return_mat_score(WHITE))
	{
		my_printf("%s\n", colour_as_string[my_colour]);
		display_move("", m, "\n");
		display_board();
		display_list();
		display_occup();
		FATAL_ERROR("do white mat error")
	}
	if (node->node_mat[BLACK] != return_mat_score(BLACK))
	{
		my_printf("%s\n", colour_as_string[my_colour]);
		display_move("", m, "\n");
		display_board();
		display_list();
		display_occup();
		FATAL_ERROR("do black mat error")
	}
#endif
#undef DO_CAPT
#undef DO_MOVE
}

void undo_move(int my_colour, int your_colour)
{
#define MOVE(A, B)\
	{\
		p = board[A];\
		p->board_where = B;\
		p->board_moved = node->node_moved;\
		p->board_captured = node->node_captured;\
		board[A] = &empty;\
		board[B] = p;\
	}

	board_t *p;
	long kind;
	int from, to;
	int ep;
#if DEBUG > 0
	int err;
#endif

	/*
	We have to peek in the previous node for the move.
	*/
	
	UNPACK_MOVE(node[-1].node_move, kind, from, to)
	if (kind == NULL_MOVE) goto label_null;
	if ((kind != SHORT) and (kind != LONG))
	{
		BUG(column[from] == INVALID)
		BUG(column[to] == INVALID)
	}
	if (kind == EP)
	{
		ep = to - pawn_dir[my_colour][0];
		p = board[ep] = node->capt;
		p->board_where = ep;
		(*p->board_count)++;
		MOVE(to, from)
		occup[your_colour][column[ep]] |= (1 << row[ep]);
		occup[my_colour][column[to]] &= ~(1 << row[to]);
		occup[my_colour][column[from]] |= (1 << row[from]);
		tot_count[your_colour]++;
	}
	else if (kind & PROM_CAPT)
	{
		p = board[to];
		p->board_kind = p->board_where = INVALID;
		(*p->board_count)--;
		
		p = board[to] = node->capt;
		p->board_where = to;
		tot_count[your_colour]++;
		(*p->board_count)++;

		p = board[from] = node->prom;
		p->board_where = from;
		(*p->board_count)++;
		occup[my_colour][column[from]] |= (1 << row[from]);
	}
	else if (kind == SHORT)
	{
		if (my_colour == WHITE)
		{
			MOVE(F1, H1)
			MOVE(G1, E1)
		}
		else
		{
			MOVE(F8, H8)
			MOVE(G8, E8)
		}
	}
	else if (kind == LONG)
	{
		if (my_colour == WHITE)
		{
			MOVE(D1, A1)
			MOVE(C1, E1)
		}
		else
		{
			MOVE(D8, A8)
			MOVE(C8, E8)
		}
	}
	else if (kind & PROM_NORM)
	{
		p = board[to];
		p->board_kind = p->board_where = INVALID;
		(*p->board_count)--;
		board[to] = &empty;
		p = board[from] = node->prom;
		p->board_where = from;
		occup[my_colour][column[from]] |= (1 << row[from]);
		(*p->board_count)++;
	}
	else
	{
		MOVE(to, from)
		if (p->board_kind == PAWN)
		{
			occup[my_colour][column[from]] |= (1 << row[from]);
			occup[my_colour][column[to]] &= ~(1 << row[to]);
		}
		if (kind == CAPT)
		{
			p = board[to] = node->capt;
			p->board_where = to;
			if (p->board_kind == PAWN)
				occup[your_colour][column[to]] |= (1 << row[to]);
			tot_count[your_colour]++;
			(*p->board_count)++;
		}
	}
	label_null:
	node--;
	inode--;
	BUG(inode < 0)
	hack_ep = node->node_hack_ep;
#if DEBUG > 0
	if ((err = check_list()) > 0)
	{
		my_printf("%s err=%d\n", colour_as_string[my_colour], err);
		display_move("", node->node_move, "\n");
		display_board();
		display_list();
		display_occup();
		FATAL_ERROR("undo list error")
	}
	if ((err = check_occup()) > 0)
	{
		my_printf("%s err=%d\n", colour_as_string[my_colour], err);
		display_move("", node->node_move, "\n");
		display_board();
		display_list();
		display_occup();
		FATAL_ERROR("undo occup error")
	}
#endif
}

/*
Score simple endgames.
*/

int return_endgame_score(int, int);
int return_endgame_score(int my_colour, int your_colour)
{
	int drow, dcol, dking;
	int ikind;
	board_t *p;
	
	BUG((tot_count[WHITE] + tot_count[BLACK]) > 3)
	drow = row[king_list[my_colour]->board_where] - row[king_list[your_colour]->board_where];
	dcol = column[king_list[my_colour]->board_where] - column[king_list[your_colour]->board_where];
	dking = ABS(drow) + ABS(dcol);
	
	if (tot_count[your_colour] == 1)
	{
		if (tot_count[my_colour] == 1)
			return(0);
		for (ikind = PAWN; ikind <= QUEEN; ikind++)
		{
			for (p = list[my_colour][ikind]; p->board_kind != INVALID; p++)
				if (p->board_where != INVALID) goto foundx;
		}
		foundx:
		if (p->board_kind == PAWN)
			return(free_pawn_grad[my_colour][p->board_where] - 0);
		else if (p->board_kind == KNIGHT)
			return(0);
		else if (p->board_kind == BISHOP)
			return(0);
		else if (p->board_kind == ROOK)
			return(SCORE_ENDGAME - dking - king_alone_grad[king_list[your_colour]->board_where]);
		else if (p->board_kind == QUEEN)
			return(SCORE_ENDGAME - dking - king_alone_grad[king_list[your_colour]->board_where]);
		else
			FATAL_ERROR("board error")
	}
	else
	{
		for (ikind = PAWN; ikind <= QUEEN; ikind++)
		{
			for (p = list[your_colour][ikind]; p->board_kind != INVALID; p++)
				if (p->board_where != INVALID) goto foundy;
		}
		foundy:
		if (p->board_kind == PAWN)
			return(0 - free_pawn_grad[your_colour][p->board_where]);
		else if (p->board_kind == KNIGHT)
			return(0);
		else if (p->board_kind == BISHOP)
			return(0);
		else if (p->board_kind == ROOK)
			return(dking + king_alone_grad[king_list[my_colour]->board_where] - SCORE_ENDGAME);
		else if (p->board_kind == QUEEN)
			return(dking + king_alone_grad[king_list[my_colour]->board_where] - SCORE_ENDGAME);
		else
			FATAL_ERROR("board error")
	}
}

/*
heapsort sorts ascending, so we add your strength and subtrect my strength.
*/

void set_strength_board(int, int);
void set_strength_board(int my_colour, int your_colour)
{
	board_t *p;
	int ikind;
	int *d, j;
	
	memcpy(strength_board + A8, attack_zero + A8, (H1 - A8 + 1) * sizeof(int));
	for (ikind = PAWN; ikind <= KING; ikind++)
	{
		for (p = list[my_colour][ikind]; p->board_kind != INVALID; p++)
		{
			if (p->board_where == INVALID) continue;
			strength_board[p->board_where] -= p->board_strength;
			if (p->board_kind == PAWN)
			{
				j = p->board_where + pawn_dir[my_colour][1];
				/*
				We always add strength to target field, empty or not.
				*/
				strength_board[j] -= p->board_strength;
				j = p->board_where + pawn_dir[my_colour][2];
				strength_board[j] -= p->board_strength;
			}
			else if (!p->board_shift)
			{
				for (d = p->board_dir; *d != 0; d++)
				{
					j = p->board_where + *d;
					strength_board[j] -= p->board_strength;
				}
			}
			else
			{
				for (d = p->board_dir; *d != 0; d++)
				{
					for (j = p->board_where + *d; board[j] == &empty; j += *d)
						strength_board[j] -= p->board_strength;
					strength_board[j] -= p->board_strength;
				}
			}
		}
		
		/* same for your colour */
		
		for (p = list[your_colour][ikind]; p->board_kind != INVALID; p++)
		{
			if (p->board_where == INVALID) continue;
			strength_board[p->board_where] += p->board_strength;
			if (p->board_kind == PAWN)
			{
				j = p->board_where + pawn_dir[your_colour][1];
				strength_board[j] += p->board_strength;
				j = p->board_where + pawn_dir[your_colour][2];
				strength_board[j] += p->board_strength;
			}
			else if (!p->board_shift)
			{
				for (d = p->board_dir; *d != 0; d++)
				{
					j = p->board_where + *d;
					strength_board[j] += p->board_strength;
				}
			}
			else
			{
				for (d = p->board_dir; *d != 0; d++)
				{
					for (j = p->board_where + *d; board[j] == &empty; j += *d)
						strength_board[j] += p->board_strength;
					strength_board[j] += p->board_strength;
				}
			}
		}
	}
}

void set_attack_board(int, int);
void set_attack_board(int my_colour, int your_colour)
{
	board_t *p;
	int ikind;
	int *d, j;
	
	memcpy(attack_board + A8, attack_zero + A8, (H1 - A8 + 1) * sizeof(int));
	for (ikind = PAWN; ikind <= KING; ikind++)
	{
		for (p = list[my_colour][ikind]; p->board_kind != INVALID; p++)
		{
			if (p->board_where == INVALID) continue;
			if (p->board_kind == PAWN)
			{
				j = p->board_where + pawn_dir[my_colour][1];
				attack_board[j]++;
				j = p->board_where + pawn_dir[my_colour][2];
				attack_board[j]++;
			}
			else if (!p->board_shift)
			{
				for (d = p->board_dir; *d != 0; d++)
				{
					j = p->board_where + *d;
					attack_board[j]++;
				}
			}
			else
			{
				for (d = p->board_dir; *d != 0; d++)
				{
					for (j = p->board_where + *d; board[j] == &empty; j += *d)
						attack_board[j]++;
					attack_board[j]++;
				}
			}
		}
		
		/* same for your colour */
		
		for (p = list[your_colour][ikind]; p->board_kind != INVALID; p++)
		{
			if (p->board_where == INVALID) continue;
			if (p->board_kind == PAWN)
			{
				j = p->board_where + pawn_dir[your_colour][1];
				attack_board[j]--;
				j = p->board_where + pawn_dir[your_colour][2];
				attack_board[j]--;
			}
			else if (!p->board_shift)
			{
				for (d = p->board_dir; *d != 0; d++)
				{
					j = p->board_where + *d;
					attack_board[j]--;
				}
			}
			else
			{
				for (d = p->board_dir; *d != 0; d++)
				{
					for (j = p->board_where + *d; board[j] == &empty; j += *d)
						attack_board[j]--;
					attack_board[j]--;
				}
			}
		}
	}
}

/*
Sets the various positional parameters. Note: the assignment truncates.
This causes no problem when optimizing evaluation function parameters.
If the best value of for example weak_queen_score would be 10, than
y[0] would converge to a value between 10.0 and 10.99 if truncation is used,
but to a value between 9.5 and 10.49 if normal rounding was used.
*/

void set_score(double *y)
{
	weak_queen_score = y[0];
	weak_rook_score = y[1];
	weak_bishop_score = y[2];
	weak_knight_score = y[3];
	isolated_pawn_score = y[4];
	king_danger_score = y[5];
	double_pawn_score = y[6];
	lagged_pawn_score = y[7];
	bishop_blocks_pawn_score = y[8];
	bishop_not_developed_score = y[9];
	knight_not_developed_score = y[10];
	weak_pawn_attack_score = y[11];
	pin_score = y[12];
	rook_on_half_open_file_score = y[13];
	knight_strong_score = y[14];
	bishop_strong_score = y[15];
	rook_on_open_file_score = y[16];
	rook_on_7th_rank_score = y[17];
	long_castle_score = y[18];
	short_castle_score = y[19];
	knight_can_move_score = y[20];
	bishop_can_move_score = y[21];
	rook_can_move_score = y[22];
}

/*
return_pawn_pos_score analyzes the pawn structure. pawn_board is set
if the pawn is weak.
*/

#if VERBOSE > 0
int return_pawn_pos_score(int, int, int);
int return_pawn_pos_score(int my_colour, int your_colour, int verbose)
#else
int return_pawn_pos_score(int, int);
int return_pawn_pos_score(int my_colour, int your_colour)
#endif
{
	board_t *p;
	int icolumn, irow;
	int sum = 0;
	int n;
	int tst;
	
	for (p = pawn_list[my_colour]; p->board_kind != INVALID; p++)
	{
		if (p->board_where == INVALID) continue;
		sum += pawn_visit[my_colour][p->board_where];
		pawn_board[p->board_where] = FALSE;
		icolumn = column[p->board_where];
		irow = row[p->board_where];
		if (bit_count[occup[my_colour][icolumn]] > 1)
		{
#if VERBOSE > 0
			if (verbose)
				my_printf("%spawn on %s is doubled\n",
					colour_as_string[my_colour], GET_NOTA(p->board_where));
#endif
			sum += double_pawn_score;
		}
		/*
		For white: we have a free pawn if the row is greater than or equal to
		the highest bit of the occupation numbers of the current, left and right row. 
		*/
		if (icolumn == 0)
			n = occup[your_colour][1];
		else if (icolumn == 7)
			n = occup[your_colour][6];
		else
			n = (occup[your_colour][icolumn - 1] | occup[your_colour][icolumn + 1]);
		n |= (occup[your_colour][icolumn] | occup[my_colour][icolumn]);
		if (my_colour == WHITE)
			tst = (irow >= bit_high[n]);
		else
			tst = (irow <= bit_low[n]);
		if (tst)
		{
#if VERBOSE > 0
			if (verbose)
				my_printf("%spawn on %s is free\n",
					colour_as_string[my_colour], GET_NOTA(p->board_where));
#endif
			sum += free_pawn_grad[my_colour][p->board_where];
			continue;
		}
		/*
		We have an isolated pawn if there are no pawns to the left and to the right.
		*/
		if (icolumn == 0)
			n = occup[my_colour][1];
		else if (icolumn == 7)
			n = occup[my_colour][6];
		else
			n = (occup[my_colour][icolumn - 1] | occup[my_colour][icolumn + 1]);
		if (n == 0)
		{
#if VERBOSE > 0
			if (verbose)
				my_printf("%spawn on %s is isolated\n",
					colour_as_string[my_colour], GET_NOTA(p->board_where));
#endif
			sum += isolated_pawn_score;
			pawn_board[p->board_where] = TRUE;
			continue;
		}
		/*
		This is just a poor check for a lagged pawn.
		*/
		if (my_colour == WHITE)
			tst = (irow < (bit_low[n] - 1));
		else
			tst = (irow > (bit_high[n] + 1));
		if (tst)
		{
#if VERBOSE > 0
			if (verbose)
				my_printf("%spawn on %s is lagged\n",
					colour_as_string[my_colour], GET_NOTA(p->board_where));
#endif
			sum += lagged_pawn_score;
			pawn_board[p->board_where] = TRUE;
		}
	}
	return(sum);
}

/*
The other pieces 
*/

#if VERBOSE > 0
int return_pos_score(int, int, int, int);
int return_pos_score(int my_colour, int your_colour, int verbose, int inverse)
#else
int return_pos_score(int, int, int);
int return_pos_score(int my_colour, int your_colour, int inverse)
#endif
{
	board_t *p, *q;
	int tst;
	int icolumn, irow;
	int n, m;
	int sum = 0;
	int j, k;
	int *d;
	
	for (p = knight_list[my_colour]; p->board_kind != INVALID; p++)
	{
		if (p->board_where == INVALID) continue;
		sum += knight_visit[my_colour][p->board_where];
		if (!p->board_moved) sum += knight_not_developed_score;
		irow = row[p->board_where];
		icolumn = column[p->board_where];
		/*
		A knight is strong if it is a 'free pawn' and covered by one or 
		(even better) two pawns
		*/
		if (icolumn == 0)
			n = occup[your_colour][1];
		else if (icolumn == 7)
			n = occup[your_colour][6];
		else
			n = (occup[your_colour][icolumn - 1] | occup[your_colour][icolumn + 1]);
		if (my_colour == WHITE)
			tst = ((n == 0) or (irow >= bit_high[n]));
		else
			tst = ((n == 0) or (irow <= bit_low[n]));
		if (tst)
		{
			q = board[p->board_where - pawn_dir[my_colour][1]];
			if ((q->board_colour == my_colour) and (q->board_kind == PAWN))
			{
#if VERBOSE > 0
				if (verbose)
					my_printf("%sknight on %s is strong\n",
						colour_as_string[my_colour], GET_NOTA(p->board_where));
#endif
				sum += knight_strong_score;
			}
			q = board[p->board_where - pawn_dir[my_colour][2]];
			if ((q->board_colour == my_colour) and (q->board_kind == PAWN))
			{
#if VERBOSE > 0
				if (verbose)
					my_printf("%sknight on %s is strong\n",
						colour_as_string[my_colour], GET_NOTA(p->board_where));
#endif
				sum += knight_strong_score;
			}
		}
		/*
		We define mobility as the number of fields to which the knight can
		move without the risk of being captured. Note: we have to check
		if attack_board >= 1, not >= 0 because the knight itself is counted
		when generating the attack board.
		*/
		n = m = 0;
		for (d = p->board_dir; *d != 0; d++)
		{
			j = p->board_where + *d;
			q = board[j];
			if (q == &empty)
			{
				if (inverse)
					tst = (attack_board[j] <= -1);
				else
					tst = (attack_board[j] >= 1);
				if (tst)
					n += knight_can_move_score;
				else
					m += knight_can_move_score;
			}
			else if (q->board_colour == your_colour)
			{
				if (q->board_kind == PAWN)
				{
					if (pawn_board[j])
					{
#if VERBOSE > 0
						if (verbose)
							my_printf("%sknight on %s attacks weak pawn on %s\n",
								colour_as_string[my_colour], GET_NOTA(p->board_where), GET_NOTA(j));
#endif
						sum += weak_pawn_attack_score;
					}
				}
			}
		}
		/*
		A knight is weak if it can only move to sqares controlled by the opponent.
		*/
		if ((n == 0) and (m > 0))
		{
#if VERBOSE > 0
			if (verbose)
				my_printf("%sknight on %s is weak\n",
					colour_as_string[my_colour], GET_NOTA(p->board_where));
#endif
			sum += weak_knight_score;
		}
		else
			sum += (n - m);
	}	
	for (p = bishop_list[my_colour]; p->board_kind != INVALID; p++)
	{
		if (p->board_where == INVALID) continue;
		if (!p->board_moved) sum += bishop_not_developed_score;
		irow = row[p->board_where];
		icolumn = column[p->board_where];
		j = p->board_where - pawn_dir[my_colour][0];
		q = board[j];
		if ((q->board_colour == my_colour) and (q->board_kind == PAWN))
		{
#if VERBOSE > 0
			if (verbose)
				my_printf("%sbishop on %s blocks pawn on %s\n",
					colour_as_string[my_colour], GET_NOTA(p->board_where), GET_NOTA(j));
#endif
			sum += bishop_blocks_pawn_score;
		}
		if (icolumn == 0)
			n = occup[your_colour][1];
		else if (icolumn == 7)
			n = occup[your_colour][6];
		else
			n = (occup[your_colour][icolumn - 1] | occup[your_colour][icolumn + 1]);
		if (my_colour == WHITE)
			tst = ((n == 0) or (irow >= bit_high[n]));
		else
			tst = ((n == 0) or (irow <= bit_low[n]));
		if (tst)
		{
			q = board[p->board_where - pawn_dir[my_colour][1]];
			if ((q->board_colour == my_colour) and (q->board_kind == PAWN))
			{
#if VERBOSE > 0
				if (verbose)
					my_printf("%sbishop on %s is strong\n",
						colour_as_string[my_colour], GET_NOTA(p->board_where));
#endif
				sum += bishop_strong_score;
			}
			q = board[p->board_where - pawn_dir[my_colour][2]];
			if ((q->board_colour == my_colour) and (q->board_kind == PAWN))
			{
#if VERBOSE > 0
				if (verbose)
					my_printf("%sbishop on %s is strong\n",
						colour_as_string[my_colour], GET_NOTA(p->board_where));
#endif
				sum += bishop_strong_score;
			}
		}
		n = m = 0;
		for (d = p->board_dir; *d != 0; d++)
		{
			for (j = p->board_where + *d; board[j] == &empty; j += *d)
			{
				if (inverse)
					tst = (attack_board[j] <= -1);
				else
					tst = (attack_board[j] >= 1);
				if (tst)
					n += bishop_can_move_score;
				else
					m += bishop_can_move_score;
			}
			q = board[j];
			if (q->board_colour == your_colour)
			{
				if (q->board_kind == PAWN)
				{
					if (pawn_board[j])
					{
#if VERBOSE > 0
						if (verbose)
							my_printf("%sbishop on %s attacks weak pawn on %s\n",
								colour_as_string[my_colour], GET_NOTA(p->board_where), GET_NOTA(j));
#endif
						sum += weak_pawn_attack_score;
					}
				}
				else
				{
					for (k = q->board_where + *d; board[k] == &empty; k += *d);
					q = board[k];
					if ((q->board_colour == your_colour) and (q->board_kind == KING))
					{
#if VERBOSE > 0
						if (verbose)
							my_printf("%sbishop on %s pins piece on %s\n",
								colour_as_string[my_colour], GET_NOTA(p->board_where), GET_NOTA(j));
#endif
						sum += pin_score;
					}
				}
			}
		}
		if ((n == 0) and (m > 0))
		{
#if VERBOSE > 0
			if (verbose)
				my_printf("%sbishop on %s is weak\n",
					colour_as_string[my_colour], GET_NOTA(p->board_where));
#endif
			sum += weak_bishop_score;
		}
		else
			sum += (n - m);
	}	
	for (p = rook_list[my_colour]; p->board_kind != INVALID; p++)
	{
		if (p->board_where == INVALID) continue;
		irow = row[p->board_where];
		icolumn = column[p->board_where];
		n = (occup[my_colour][icolumn] | occup[your_colour][icolumn]);
		if (n == 0)
		{
#if VERBOSE > 0
			if (verbose)
				my_printf("%srook on %s is on open file\n",
					colour_as_string[my_colour], GET_NOTA(p->board_where));
#endif
			sum += rook_on_open_file_score;
		}
		else if ((occup[my_colour][icolumn] == 0) or 
		         (occup[your_colour][icolumn] == 0))
		{
#if VERBOSE > 0
			if (verbose)
				my_printf("%srook on %s is on half open file\n",
					colour_as_string[my_colour], GET_NOTA(p->board_where));
#endif
			sum += rook_on_half_open_file_score;
		}
		if (irow == strong_row[my_colour])
		{
#if VERBOSE > 0
			if (verbose)
				my_printf("%srook on %s is on 7th rank\n",
					colour_as_string[my_colour], GET_NOTA(p->board_where));
#endif
			sum += rook_on_7th_rank_score;
		}
		n = m = 0;
		for (d = p->board_dir; *d != 0; d++)
		{
			for (j = p->board_where + *d; board[j] == &empty; j += *d)
			{
				if (inverse)
					tst = (attack_board[j] <= -1);
				else
					tst = (attack_board[j] >= 1);
				if (tst)
					n += rook_can_move_score;
				else
					m += rook_can_move_score;
			}
			q = board[j];
			if (q->board_colour == your_colour)
			{
				if (q->board_kind == PAWN)
				{
					if (pawn_board[j])
					{
#if VERBOSE > 0
						if (verbose)
							my_printf("%srook on %s attacks weak pawn on %s\n",
								colour_as_string[my_colour], GET_NOTA(p->board_where), GET_NOTA(j));
#endif
						sum += weak_pawn_attack_score;
					}
				}
				else
				{
					for (k = q->board_where + *d; board[k] == &empty; k += *d);
					q = board[k];
					if ((q->board_colour == your_colour) and (q->board_kind == KING))
					{
#if VERBOSE > 0
						if (verbose)
							my_printf("%srook on %s pins piece on %s\n",
								colour_as_string[my_colour], GET_NOTA(p->board_where), GET_NOTA(j));
#endif
						sum += pin_score;
					}
				}
			}
		}
		if ((n == 0) and (m > 0))
		{
#if VERBOSE > 0
			if (verbose)
				my_printf("%sweak rook on %s\n",
					colour_as_string[my_colour], GET_NOTA(p->board_where));
#endif
			sum += weak_rook_score;
		}
		else
			sum += (n - m);
	}
	for (p = queen_list[my_colour]; p->board_kind != INVALID; p++)
	{
		if (p->board_where == INVALID) continue;
		n = m = 0;
		for (d = p->board_dir; *d != 0; d++)
		{
			for (j = p->board_where + *d; board[j] == &empty; j += *d)
			{
				if (inverse)
					tst = (attack_board[j] <= -1);
				else
					tst = (attack_board[j] >= 1);
				if (tst)
					n++;
				else
					m++;
			}
			q = board[j];
			if (q->board_colour == your_colour)
			{
				if (q->board_kind == PAWN)
				{
					if (pawn_board[j])
					{
#if VERBOSE > 0
						if (verbose)
							my_printf("%squeen on %s attacks weak pawn on %s\n",
								colour_as_string[my_colour], GET_NOTA(p->board_where), GET_NOTA(j));
#endif
						sum += weak_pawn_attack_score;
					}
				}
				else
				{
					for (k = q->board_where + *d; board[k] == &empty; k += *d);
					q = board[k];
					if ((q->board_colour == your_colour) and (q->board_kind == KING))
					{
#if VERBOSE > 0
						if (verbose)
							my_printf("%squeen on %s pins piece on %s\n",
								colour_as_string[my_colour], GET_NOTA(p->board_where), GET_NOTA(j));
#endif
						sum += pin_score;
					}
				}
			}
		}
		if ((n == 0) and (m > 0))
		{
#if VERBOSE > 0
			if (verbose)
				my_printf("%sweak queen on %s\n",
					colour_as_string[my_colour], GET_NOTA(p->board_where));
#endif
			sum += weak_queen_score;
		}
	}
	/*
	King safety
	*/
	p = king_list[my_colour];
	for (d = p->board_dir; *d != 0; d++)
	{
		j = p->board_where + *d;
		if (board[j] == &side) continue;
		if (inverse)
			tst = (attack_board[j] > 0);
		else
			tst = (attack_board[j] < 0);
		if (tst)
		{
#if VERBOSE > 0
			if (verbose)
				my_printf("%sking on %s is in danger on %s\n",
					colour_as_string[my_colour], GET_NOTA(p->board_where), GET_NOTA(j));
#endif
			sum += king_danger_score;
		}
	}
	return(sum);
}

void display_score(void)
{
	int nwhite_mat, nblack_mat;
	int nwhite_pawn, nblack_pawn;
	int nwhite_pos, nblack_pos;
	int nwhite_tot, nblack_tot;
	int ndiff, nendgame;
	
	if ((tot_count[WHITE] + tot_count[BLACK]) < 4)
	{
		nendgame = return_endgame_score(WHITE, BLACK);
		my_printf("endgame=%d\n", nendgame);
		return;
	}
	
	set_attack_board(WHITE, BLACK);
	nwhite_mat = return_mat_score(WHITE);
	nblack_mat = return_mat_score(BLACK);
#if VERBOSE > 0
	nwhite_pawn = return_pawn_pos_score(WHITE, BLACK, TRUE);
	nblack_pawn = return_pawn_pos_score(BLACK, WHITE, TRUE);
	nwhite_pos = return_pos_score(WHITE, BLACK, TRUE, FALSE);
	nblack_pos = return_pos_score(BLACK, WHITE, TRUE, TRUE);
#else
	nwhite_pawn = return_pawn_pos_score(WHITE, BLACK);
	nblack_pawn = return_pawn_pos_score(BLACK, WHITE);
	nwhite_pos = return_pos_score(WHITE, BLACK, FALSE);
	nblack_pos = return_pos_score(BLACK, WHITE, TRUE);
#endif
	nwhite_tot = nwhite_pawn + nwhite_pos;		
	nblack_tot = nblack_pawn + nblack_pos;

	my_printf("mat score white=%d black=%d\n", nwhite_mat, nblack_mat);
	my_printf("pawn score white=%d black=%d\n", nwhite_pawn, nblack_pawn);
	my_printf("pos score white=%d black=%d\n", nwhite_pos, nblack_pos);
	my_printf("tot pos score white=%d black=%d\n", nwhite_tot, nblack_tot);
	ndiff = nwhite_mat - nblack_mat + nwhite_tot - nblack_tot;
	my_printf("diff score=%d\n", ndiff);
}

/*
We check for draw by repetition by checking if the current key has occurred
before. In the alpha_beta loop, we return TRUE when we find a match.
In the main loop, we must check whether the position has occurred three times.
*/

int draw_by_repetition(int strict)
{
	node_t *temp;
	int ncount;
	
  temp = node;
  ncount = 1;
  while(TRUE)
  {
  	/*
  	your move and my move must be reversible.
  	*/
  	if (!temp->node_reverse) break;
  	temp--;
  	if (!temp->node_reverse) break;
  	temp--;
  	if ((node->node_total_key.lo == temp->node_total_key.lo) and
  	    (node->node_total_key.hi == temp->node_total_key.hi))
  	{
  		ncount++;
  		if (!strict)
  			return(TRUE);
  		else if (ncount > 2)
  			return(TRUE);
  	}
  }
  return(FALSE);
}

/*
heapsort routine. Sorts the move strength a and the moves m.
A straight implementation in C of the algorithm described by Knuth in
The art of computer programming, Volume 3, pages 146-147.
The arrays should be offset one.
*/

void heapsort(int n, int *, move_t *);
void heapsort(int n, int *a, move_t *m)
{
	int l, r, b, i, j;
	move_t t;
	
	if (n < 2) return;
	l = n / 2 + 1;
	r = n;
	h2:
	if (l > 1)
	{
		l--;
		b = a[l]; t = m[l];
	}
	else
	{
		b = a[r]; t = m[r];
		a[r] = a[1]; m[r] = m[1];
		r--;
		if (r == 1)
		{
			a[1] = b; m[1] = t;
			goto h9;
		}		
	}
	j = l;
	h4:
	i = j;
	j = 2 * j;
	if (j == r) goto h6;
	if (j > r) goto h8;
	if (a[j] < a[j + 1]) j++;
	h6:
	if (b >= a[j]) goto h8;
	a[i] = a[j]; m[i] = m[j];
	goto h4;
	h8:
	a[i] = b; m[i] = t;
	goto h2;
	h9:
	for (i = 1; i < n; i++)
		BUG(a[i] > a[i + 1])
}

#define QUIES_DEPTH 2

/*
A difficult problem in any chess program is the quiescense search (the capture
evaluation). See for example
G. Schrufer, ICCA vol. 12, no. 1 (1989) pp. 3-9.
I do it in the following way: any piece that has captured is marked.
In the quiescense search, I allow QUIES_DEPTH free captures (so any piece
can capture any other piece), but after the QUIES_DEPTH free captures, only
captures to pieces which have their capture mark set are allowed. In this way the
quiescence search will always terminate. Of course higher QUIES_DEPTH values
give better results, but also makes the program slower, so there is a trade-off
here.
*/

int quies(int, int, int, int, int, int);
int quies(int swap, int my_colour, int your_colour, int alpha, int beta, int depth)
{
	int temp;
	int best_score;
	score_cache_t *p;
	int nmoves;
	int i;
	int ikind;
	board_t *q;
	move_t best_move = INVAL_MOVE;
	quies_cache_t *a;
	
	if (swap)
	{
		temp = alpha;
		alpha = -beta;
		beta = -temp;
	}
	
	node->capt_base = capt_top;

	best_score = node->node_mat[my_colour] - node->node_mat[your_colour] + 
	             node->node_pos[my_colour] - node->node_pos[your_colour];
	if ((tot_count[WHITE] + tot_count[BLACK]) < 4)
	{
		best_score += return_endgame_score(my_colour, your_colour);
		goto label_neg;
	}
	p = score_cache[my_colour] + (node->node_total_key.lo & SCORE_CACHE_MASK);
	if ((p->cache_score != SCORE_INVAL) and
	    (p->cache_key.lo == node->node_total_key.lo) and
	    (p->cache_key.hi == node->node_total_key.hi))
	{
		best_score += p->cache_score;
		nhits++;
	}
	else
	{
		set_attack_board(my_colour, your_colour);
#if VERBOSE > 0
		temp =
			return_pawn_pos_score(my_colour, your_colour, FALSE) -
			return_pawn_pos_score(your_colour, my_colour, FALSE);
		temp +=
			return_pos_score(my_colour, your_colour, FALSE, FALSE) -
			return_pos_score(your_colour, my_colour, FALSE, TRUE);
#else
		temp =
			return_pawn_pos_score(my_colour, your_colour) -
			return_pawn_pos_score(your_colour, my_colour);
		temp +=
			return_pos_score(my_colour, your_colour, FALSE) -
			return_pos_score(your_colour, my_colour, TRUE);
#endif
		nevals++;
		best_score += temp;
		p->cache_score = temp;
		p->cache_key = node->node_total_key;
	}
	
	/*
	If by doing nothing we already have a fail high, we do not have to 
	check the captures.
	*/
	
	if (best_score >= beta)
		goto label_neg;

	/*
	The captures. We generate the captures to your most valuable pieces first.
	*/
	for (ikind = QUEEN; ikind >= PAWN; ikind--)
	{
		for (q = list[your_colour][ikind]; q->board_kind != INVALID; q++)
		{
			if (q->board_where == INVALID) continue;
			/*
			Only recaptures if (depth >= QUIES_DEPTH)
			*/
			if ((depth >= QUIES_DEPTH) and !q->board_captured) continue;
			capt_top = node->capt_base;
			nmoves = gen_capts(q->board_where, my_colour);
			for (i = 0; i < nmoves; i++)
			{
				do_move(my_colour, your_colour, node->capt_base[i]);
				if (attacked(king_list[my_colour]->board_where, your_colour))
				{
					undo_move(my_colour, your_colour);
					continue;
				}
				temp = quies(TRUE, your_colour, my_colour, MAX(alpha, best_score), beta, depth + 1);
				undo_move(my_colour, your_colour);
				if (temp > best_score)
				{
					best_score = temp;
					best_move = node->capt_base[i];
					if (best_score >= beta) goto label_neg;
				}
			}
		}
	}
	
	label_neg:
	/*
	Store the best move in the quies_cache.
	*/
	if (best_move != INVAL_MOVE)
	{
		a = quies_cache[my_colour] + (node->node_total_key.lo & QUIES_CACHE_MASK);
		a->quies_move = best_move;
		a->quies_key = node->node_total_key;
	}
	capt_top = node->capt_base;
	if (swap)	
		return(-best_score);
	else
		return(best_score);
}

/*
Because we only use 64 bit hash keys, hash faults can occur (different positions
with the same hash key). The program will crash when doing the best
move from the hash table in that position. So we have to check if the best move
from the hash table is a legal move. As we do not want to do a move generation,
we only check if the moves is legal in the sense that no data structures 
are destroyed or illegal positions are reached. E.g. we do not check if square 
F1 is attacked if the best move is short castle.
*/

int legal_move(int, int, move_t);
int legal_move(int my_colour, int your_colour, move_t m)
{
	long kind;
	int from, to;
	board_t *p;
	
	UNPACK_MOVE(m, kind, from, to)
	if (kind == NULL_MOVE) return(TRUE);
	if (kind == EP)
	{
		p = board[from];
		if ((p->board_colour != my_colour) or (p->board_kind != PAWN)) return(FALSE);
		p = board[to];
		if (p != &empty) return(FALSE);
		p = board[to - pawn_dir[my_colour][0]];
		if ((p->board_colour != your_colour) or (p->board_kind == KING)) return(FALSE);
	}
	else if (kind & PROM_CAPT)
	{
		p = board[from];
		if (p->board_colour != my_colour) return(FALSE);
		if (p->board_kind != PAWN) return(FALSE);
		p = board[to];
		if ((p->board_colour != your_colour) or (p->board_kind == KING)) return(FALSE);
	}
	else if (kind & PROM_NORM)
	{
		p = board[from];
		if (p->board_colour != my_colour) return(FALSE);
		if (p->board_kind != PAWN) return(FALSE);
		p = board[to];
		if (p != &empty) return(FALSE);
	}
	else if (kind == SHORT)
	{
		if (my_colour == WHITE)
		{
			if (board[E1]->board_moved) return(FALSE);
			if (board[H1]->board_moved) return(FALSE);
			if (board[F1] != &empty) return(FALSE);
			if (board[G1] != &empty) return(FALSE);
		}
		else
		{
			if (board[E8]->board_moved) return(FALSE);
			if (board[H8]->board_moved) return(FALSE);
			if (board[F8] != &empty) return(FALSE);
			if (board[G8] != &empty) return(FALSE);
		}
	}
	else if (kind == LONG)
	{
		if (my_colour == WHITE)
		{
			if (board[E1]->board_moved) return(FALSE);
			if (board[A1]->board_moved) return(FALSE);
			if (board[D1] != &empty) return(FALSE);
			if (board[C1] != &empty) return(FALSE);
			if (board[B1] != &empty) return(FALSE);
		}
		else
		{
			if (board[E8]->board_moved) return(FALSE);
			if (board[A8]->board_moved) return(FALSE);
			if (board[D8] != &empty) return(FALSE);
			if (board[C8] != &empty) return(FALSE);
			if (board[B8] != &empty) return(FALSE);
		}
	}
	else if (kind == CAPT)
	{
		p = board[from];
		if (p->board_colour != my_colour) return(FALSE);
		p = board[to];
		if ((p->board_colour != your_colour) or (p->board_kind == KING)) return(FALSE);
	}
	else
	{
		p = board[from];
		if (p->board_colour != my_colour) return(FALSE);
		p = board[to];
		if (p != &empty) return(FALSE);
	}
	return(TRUE);
}

/*
A standard alpha_beta routine, but with the null move heuristic, strength
heuristic and threat extensions. See for example
T.A. Marsland, ICCA vol. 9, no. 1 (1986) pp. 3-19.
For the guard heuristic see
ICCA vol. 16, no. 2 (1993).
For the null move heuristic see 
Chr. Donninger, ICCA vol. 9, no 3 (1993) pp. 137-143.
For the threat extensions see
T.S. Ananthamaran, ICCA vol. 14, no. 2 (1991) pp. 47-65.
Null moves are used as a forward pruning mechanism: if the null move gives
a beta cutoff, it is assumed that the regular moves in that position, when searched,
would also give a beta cutoff. This speeds up the program by a factor of two. but
it fails terribly in case of zugzwang, because in that case not doing a move is a 
usually a good idea: the null move gives a beta cutuff, but the regular moves will not.
I call this 'null move failures'. To guard against these, I only allow
null moves when we are not in the principle variation. This protects against
incorrect fail high's in the aspiration search, but not against incorrect fail lows.
It is very easy to check how often null move failures occur: instead of
doing the null move cutoff, you set a flag. If you do not get a beta cutoff
when searching the regular moves and that flag is set, you know you have 
a null move failure. I tried this for the 300 win-at-chess positions 
from Lang&Smith, searched with DepthMax = 5. In 5 out of the 300 positions,
I got null move failure warnings. It still found the right solution
in 4 of the 5 positions, and in 1 position it did not find the solution any more, 
because the best move failed low in the aspiration search.
Threat extensions are used to extend fail high moves by one more ply if 
there is a 'threat'. There is a threat if the null move score is a certian margin
(I use 2.5 pawn) below the current material balance. Unlike the null move
heuristic, threat extensions are 'safe': you can only get incorrect extensions
in cases of zugzwang.
NOTE! The threat extensions of Ananthamaran depend on alpha.
This is a BAD idea. Moves that failed high, may fail low when researched
(alpha changed, so some threat extensions may not occur),
making it impossible to determine the true score of the move.
*/

int alpha_beta(int, int, int, int, int);
int alpha_beta(int my_colour, int your_colour, int alpha, int beta, int depth)
{
#define RETURN(X)\
	{best_score = (X); goto label_neg;}
	
	/*
	These macro's were simple in previous versions of the program.
	GET_THREAT checks for a null move threat.
	ANAL_MOVE takes a move and analyzes it. If the move failes high,
	it is extended if there is a threat.
	*/
	
#define GET_THREAT\
	if (threat_score == SCORE_INVAL)\
	{\
		null_move = TRUE;\
		do_move(my_colour, your_colour, NULL_MOVE);\
		depth_next = depth - 1;\
		threat_score = alpha_beta(your_colour, my_colour,\
			mat_diff - THREAT_MARGIN - 1, mat_diff - THREAT_MARGIN,\
			depth_next - THREAT_REDUC);\
		undo_move(my_colour, your_colour);\
		null_move = FALSE;\
		if (threat_score == SCORE_INVAL) return(SCORE_INVAL);\
	}
#define ANAL_MOVE(M)\
	if (found and ((M) == skip_move)) continue;\
	do_move(my_colour, your_colour, (M));\
	if (attacked(king_list[my_colour]->board_where, your_colour))\
	{\
		undo_move(my_colour, your_colour);\
		continue;\
	}\
	nlegal++;\
	ndos++;\
	temp = alpha_beta(your_colour, my_colour, MAX(alpha, best_score), beta, depth_next);\
	undo_move(my_colour, your_colour);\
	if (temp == SCORE_INVAL) return(SCORE_INVAL);\
	if (temp >= beta)\
	{\
		if (!check and !null_move and\
		    (nthreats < NTHREAT_MAX) and (depth >= THREAT_DEPTH))\
		{\
			GET_THREAT\
			if ((threat_score + THREAT_MARGIN) <= mat_diff)\
			{\
				nextends++;\
				do_move(my_colour, your_colour, (M));\
				nthreats++;\
				temp = alpha_beta(your_colour, my_colour, alpha, beta, depth_next + 1);\
				nthreats--;\
				undo_move(my_colour, your_colour);\
				if (temp == SCORE_INVAL) return(SCORE_INVAL);\
			}\
		}\
	}\
	if (temp > best_score)\
	{\
		best_score = temp;\
		best_move = (M);\
		if (best_score >= beta) goto label_store;\
	}

	int temp;
	cache_t *p;
	int best_score;
	move_t best_move = INVAL_MOVE, skip_move;
	int check;
	int nmoves, nlegal;
	int i;
	int depth_next;
	int k;
	board_t *q;
	int found;
	int threat_score = SCORE_INVAL;
	int mat_diff = node->node_mat[my_colour] - node->node_mat[your_colour];
	
#if DEBUG > 0
	node_t *node_debug = node;
#endif
#if MACHINE == MAC
	if (Button()) return(SCORE_INVAL);
#endif
	
	BUG(alpha == beta)
	BUG(depth < 0)
	
	/*
	A previous version of my program occasionally stuck
	in infinite loops when run on an RS6000. After some debugging,
	I found out that it had to do with the time control:
	The code looked something like:
	t1 = clock();
	..
	t2 = clock();
	if ((t2 - t1) > (TimeLimit * CLOCKS_PER_SEC)) return..
	
	Looks good, but only works on machines with a slow clock (like my MAC).
	On an RS6000 CLOCKS_PER_SEC is 1000000L, and clock_t is a long,
	so the clock cycles in about 40 minutes. Now suppose t1 is LONG_MAX (or
	near LONG_MAX), then t2 is always less than t1 and the program
	will not terminate!
	To overcome this problem, I now update the time used every 1/10 of 
	a second. When a clock cycle is detected a warning is printed (no attempt
	is made to recover from the clock cycle, so the time used may by off
	by at most 1/10 of a second.
	*/
	
	clock_end = clock();
	if (clock_end < clock_begin)
	{
		my_printf("clock time cycle detected!\n");
		clock_begin = clock_end;
	}
	else
	{
		if ((clock_end - clock_begin) > (CLOCKS_PER_SEC / 10))
		{
			time_used += (double) (clock_end - clock_begin) / CLOCKS_PER_SEC;
			clock_begin = clock_end;
			if (time_used > TimeLimit) return(SCORE_INVAL);
		}
	}
	
	/*
	I swap alpha and beta here to keep the call to alpha_beta readable.
	*/
	
	temp = alpha;
	alpha = -beta;
	beta = -temp;
	
	node->capt_base = capt_top;
	node->move_base = move_top;

	nodes++;
	
	best_score = -SCORE_LIMIT;
	
	/*
	We first check the hash table. Since we use the lower CACHE_KIND_BITS bits
	to store some information, we do not check these bits.
	*/
	
	found = FALSE;
	p = cache[my_colour] + (node->node_total_key.lo & CACHE_MASK);
	if (!(p->cache_key.lo & CACHE_INVAL) and
	    (p->cache_key.hi == node->node_total_key.hi) and
	    ((p->cache_key.lo & ~CACHE_KIND_MASK) == (node->node_total_key.lo & ~CACHE_KIND_MASK)))
	{
		found = TRUE;
		best_move = p->cache_best;
		
		/*
		Checking if cache_depth >= depth is possibly dangerous.
		*/
		if ((p->cache_key.lo & CACHE_DEPTH_MASK) >= depth)
		{
			if (p->cache_key.lo & CACHE_TRUE)
				RETURN(p->cache_score)
			else if (p->cache_key.lo & CACHE_LOWER)
			{
				if (p->cache_score >= beta)
					RETURN(p->cache_score)
				else if (p->cache_score > alpha)
					alpha = p->cache_score;
			}
			else if (p->cache_key.lo & CACHE_UPPER)
			{
				if (p->cache_score <= alpha)
					RETURN(p->cache_score)
				else if (p->cache_score < beta)
					beta = p->cache_score;
			}
		}
	}
	
	if (draw_by_repetition(FALSE)) RETURN(0)
	
	/*
	king_list[my_colour][0].board_where for the unspoilt C-programmer..
	*/
	
	check = attacked(king_list[my_colour]->board_where, your_colour);
	
	/*
	We have an end node if we are not check and depth == 0.
	*/
	
	if (!check and (depth == 0))
	{
		best_score = quies(FALSE, my_colour, your_colour, alpha, beta, 0);
		goto label_neg;
	}

	/*
	Try the null move. We assume no zugzwang if material >= (2*ROOK_SCORE).
	*/
	
	if (!check and !principle_variation and !null_move and
	    (depth >= NULL_MOVE_DEPTH) and (node->node_mat[my_colour] >= (2*ROOK_SCORE)))
	{
		null_move = TRUE;
		do_move(my_colour, your_colour, NULL_MOVE);
		depth_next = depth - 1;
		temp = alpha_beta(your_colour, my_colour, beta - 1, beta,
			depth_next - NULL_MOVE_REDUC);
		undo_move(my_colour, your_colour);
		null_move = FALSE;
		if (temp == SCORE_INVAL) return(SCORE_INVAL);
		if (temp >= beta) RETURN(temp)
	}

	/*
	Since we generate pseudo legal moves, we have to keep track of 
	the number of legal moves.
	*/
	
	nlegal = 0;
	depth_next = (check ? depth : depth - 1);
	
	/*
	First we try the best move from the hash table. We check for hash faults.
	*/
	
	if (found and (best_move != INVAL_MOVE))
	{
		if (!legal_move(my_colour, your_colour, best_move))
		{
			my_printf("hash fault, illegal move!\n");
			display_board();
			display_move(colour_as_string[my_colour], best_move, "\n");
		}
		else
		{
			do_move(my_colour, your_colour, best_move);
			if (attacked(king_list[my_colour]->board_where, your_colour))
			{
				undo_move(my_colour, your_colour);
				my_printf("hash fault, check!\n");
				display_board();
				display_move(colour_as_string[my_colour], best_move, "\n");
			}
			else
			{
				nlegal++;
				ndos++;
				best_score = alpha_beta(your_colour, my_colour, alpha, beta, depth_next);
				undo_move(my_colour, your_colour);
				if (best_score == SCORE_INVAL) return(SCORE_INVAL);
				if (best_score >= beta)
				{
					if (!check and !null_move and
					    (nthreats < NTHREAT_MAX) and (depth >= THREAT_DEPTH))
					{
						GET_THREAT
						if ((threat_score + THREAT_MARGIN) <= mat_diff)
						{
							nextends++;
							do_move(my_colour, your_colour, best_move);
							nthreats++;
							best_score = alpha_beta(your_colour, my_colour, alpha, beta, depth_next + 1);
							nthreats--;
							undo_move(my_colour, your_colour);
							if (best_score == SCORE_INVAL) return(SCORE_INVAL);
						}
					}
				}
				if (best_score >= beta)
					goto label_store;
			}
		}
	}

	ngens++;
	
	/*
	I recently implemented the strength heuristic. Saves about 10%
	on the winatchess positions. Use #undef USE_STRENGTH to use the old
	version.
	*/

#ifdef USE_STRENGTH
	set_strength_board(my_colour, your_colour);
	capt_strength = capt_strength_table;
	move_strength = move_strength_table;
	
	/*
	Generate all moves
	*/
	
	for (k = PAWN; k <= KING; k++)
	{
		for (q = list[my_colour][k]; q->board_kind != INVALID; q++)
		{
			if (q->board_where == INVALID) continue;
			nmoves = gen_moves(my_colour, your_colour, q);
		}
	}
	
	/*
	Sort the captures.
	*/
	
	heapsort(nmoves, capt_strength_table - 1, node->capt_base - 1);
	nmoves = gen_special_moves(my_colour, your_colour);
	
	/*
	Sort the non-capturing moves.
	*/
	
	nmoves = move_top - node->move_base;
	heapsort(nmoves, move_strength_table - 1, node->move_base - 1);
	
	/*
	Search the captures and special moves.
	*/
	
	nmoves = capt_top - node->capt_base;
	skip_move = best_move;
	for (i = 0; i < nmoves; i++)
	{
		ANAL_MOVE(node->capt_base[i])
	}
	
	/*
	Search the non-capturing moves.
	*/
	
	nmoves = move_top - node->move_base;
	for (i = 0; i < nmoves; i++)
	{
		ANAL_MOVE(node->move_base[i])
	}
#else
	skip_move = best_move;
	
	/*
	Generate the captures 'piece by piece' and search them.
	*/
	for (k = PAWN; k <= KING; k++)
	{
		for (q = list[my_colour][k]; q->board_kind != INVALID; q++)
		{
			if (q->board_where == INVALID) continue;
			capt_top = node->capt_base;
			capt_strength = capt_strength_table;
			move_strength = move_strength_table;
			nmoves = gen_moves(my_colour, your_colour, q);
			for (i = 0; i < nmoves; i++)
			{
				ANAL_MOVE(node->capt_base[i])
			}
		}
	}
	
	/*
	Search the special moves.
	*/
	
	capt_top = node->capt_base;
	nmoves = gen_special_moves(my_colour, your_colour);
	for (i = 0; i < nmoves; i++)
	{
		ANAL_MOVE(node->capt_base[i])
	}
	
	/*
	Finally the non-capturing moves.
	*/
	
	nmoves = move_top - node->move_base;
	for (i = 0; i < nmoves; i++)
	{
		ANAL_MOVE(node->move_base[i])
	}
#endif
	
	/*
	When detecting check mate, we add inode to the mate score
	to make sure we find the shortest mate.
	*/
	
	if (nlegal == 0)
	{
		if (check)
			RETURN(-SCORE_MATE + inode)
		else
			RETURN(0)
	}
	
	label_store:
	
	/*
	We do not overwrite PV entries when we are not in the 
	principle variation.
	*/
	
	if ((p->cache_key.lo & CACHE_PV) and !principle_variation)
		goto label_neg;
	p->cache_key.hi = node->node_total_key.hi;
	
	/*
	We stuff the depth and the flags in the lower CACHE_KIND_BITS.
	*/
	
	p->cache_key.lo = ((node->node_total_key.lo & ~CACHE_KIND_MASK) | depth);
	p->cache_score = best_score;
	if (best_score <= alpha)
		p->cache_key.lo |= CACHE_UPPER;
	else if (best_score >= beta)
		p->cache_key.lo |= CACHE_LOWER;
	else
		p->cache_key.lo |= CACHE_TRUE;
	if (principle_variation)
		p->cache_key.lo |= CACHE_PV;
	p->cache_best = best_move;
	
	label_neg:
	BUG(best_score == SCORE_INVAL)
#if DEBUG > 0
	BUG(node != node_debug)
#endif
	capt_top = node->capt_base;
	move_top = node->move_base;
	return(-best_score);
#undef RETURN
#undef GET_THREAT
#undef ANAL_MOVE
}

/*
The legal moves in the starting position are stored in array pv.
*/

int npv;
pv_t pv[NPVS_MAX];
int jpv;

/*
gen_pv finds the legal moves in the starting position.
Gives a nice example of copy/paste coding.
*/

void gen_pv(int my_colour, int your_colour)
{
	int i, nmoves;
	int ikind;
	board_t *p;
	int check;
				
	npv = 0;
	capt_top = capt_stack;
	move_top = move_stack;
	node->capt_base = capt_top;
	node->move_base = move_top;
	for (ikind = PAWN; ikind <= KING; ikind++)
	{
		for (p = list[my_colour][ikind]; p->board_kind != INVALID; p++)
		{
			if (p->board_where == INVALID) continue;
			capt_top = node->capt_base;
			capt_strength = capt_strength_table;
			move_strength = move_strength_table;
			nmoves = gen_moves(my_colour, your_colour, p);
			for (i = 0; i < nmoves; i++)
			{
				do_move(my_colour, your_colour, node->capt_base[i]);
				check = attacked(king_list[my_colour]->board_where, your_colour);
				undo_move(my_colour, your_colour);
				if (!check)
				{
					BUG(npv >= NPVS_MAX)
					pv[npv].pv_move = node->capt_base[i];
					npv++;
				}
			}
		}
	}
	capt_top = node->capt_base;
	capt_strength = capt_strength_table;
	move_strength = move_strength_table;
	nmoves = gen_special_moves(my_colour, your_colour);
	for (i = 0; i < nmoves; i++)
	{
		do_move(my_colour, your_colour, node->capt_base[i]);
		check = attacked(king_list[my_colour]->board_where, your_colour);
		undo_move(my_colour, your_colour);
		if (!check)
		{
			BUG(npv >= NPVS_MAX)
			pv[npv].pv_move = node->capt_base[i];
			npv++;
		}
	}

	nmoves = move_top - node->move_base;
	for (i = 0; i < nmoves; i++)
	{
		do_move(my_colour, your_colour, node->move_base[i]);
		check = attacked(king_list[my_colour]->board_where, your_colour);
		undo_move(my_colour, your_colour);
		if (!check)
		{
			BUG(npv >= NPVS_MAX)
			pv[npv].pv_move = node->move_base[i];
			npv++;
		}
	}
	for (i = 0; i < npv; i++)
	{
		move_to_string(pv[i].pv_move, pv[i].pv_string);
		pv[i].pv_score = 0;
		pv[i].pv_hit = 0;
	}
	jpv = 0;
}

/*
search_pv checks whether move occurs in the list of pv moves.
Returns INVALID if the move is not found.
Prints the legal if the move is not found and the verbose flag is set.
*/

int search_pv(char *move, int verbose)
{
	int ipv;
	
	for (ipv = 0; ipv < npv; ipv++)
		if (strcmp(move, pv[ipv].pv_string) == 0) break;
	if (ipv < npv) return(ipv);
	if (verbose)
	{
		display_board();
		my_printf("move=%s\n", move);
		for (ipv = 0; ipv < npv; ipv++)
			my_printf("%s\n", pv[ipv].pv_string);
	}
	return(INVALID);
}

/*
analyze_cache tries to find the principle variation in the hash tables.
best_move is the best move in the current position.
*/

void analyze_cache(int, int, move_t);
void analyze_cache(int my_colour, int your_colour, move_t best_move)
{
	cache_t *p;
	quies_cache_t *a;
	
	display_move(" ", best_move, "");
	if (draw_by_repetition(FALSE))
	{
		my_printf(" rep!");
		return;
	}
	do_move(my_colour, your_colour, best_move);
	if (attacked(king_list[my_colour]->board_where, your_colour))
	{
		my_printf("hash fault, check!\n");
		display_board();
		display_move(colour_as_string[my_colour], best_move, "\n");
	}
	else
	{
		/*
		First we check the main hash table.
		*/
		
		p = cache[your_colour] + ((node->node_total_key.lo) & CACHE_MASK);
		if ((p->cache_key.lo & CACHE_PV) and
		    (p->cache_key.hi == node->node_total_key.hi) and 
		    ((p->cache_key.lo & ~CACHE_KIND_MASK) == (node->node_total_key.lo & ~CACHE_KIND_MASK)))
		{
			if (!legal_move(your_colour, my_colour, p->cache_best))
			{
				my_printf("hash fault, illegal move!\n");
				display_board();
				display_move(colour_as_string[your_colour], p->cache_best, "\n");
			}
			else
				analyze_cache(your_colour, my_colour, p->cache_best);
		}
		else
		{
			/*
			Then we try the quiescence hash table.
			*/
			a = quies_cache[your_colour] + ((node->node_total_key.lo) & QUIES_CACHE_MASK);
			if ((a->quies_move != INVAL_MOVE) and
				  (a->quies_key.hi == node->node_total_key.hi) and
				  (a->quies_key.lo == node->node_total_key.lo))
			{
				if (!legal_move(your_colour, my_colour, a->quies_move))
				{
					my_printf("hash fault, illegal capt!\n");
					display_board();
					display_move(colour_as_string[your_colour], a->quies_move, "\n");
				}
				else
					analyze_cache(your_colour, my_colour, a->quies_move);
			}
		}
	}
	undo_move(my_colour, your_colour);
}

void clear_cache(void);
void clear_cache(void)
{
	long i;

	for (i = 0; i < (CACHE_MASK + 1); i++)
		cache[WHITE][i].cache_key.lo = cache[BLACK][i].cache_key.lo = CACHE_INVAL;
	for (i = 0; i < (SCORE_CACHE_MASK + 1); i++)
		score_cache[WHITE][i].cache_score = score_cache[BLACK][i].cache_score = SCORE_INVAL;
	for (i = 0; i < (QUIES_CACHE_MASK + 1); i++)
		quies_cache[WHITE][i].quies_move = quies_cache[BLACK][i].quies_move = INVAL_MOVE;
}

/*
get_book_move checks whether moves are in the opening book. Captures
first (to prevent the program from delaying recaptures), than non-capturing
moves.
*/
int get_book_move(int my_colour, int your_colour)
{
	int temp[NPVS_MAX];
	int ntemp, itemp, ipv, ibook;
	book_t *with;
	int hit_max;
	int has_capt;
	
	if (!UseBook) return(INVALID);

	/*
	ntemp is the number of book moves found.
	hit_max is the number of variations in which the move occurs.
	*/
	
	ntemp = 0;
	hit_max = 0;
	has_capt = FALSE;
	for (ipv = 0; ipv < npv; ipv++)
	{
		do_move(my_colour, your_colour, pv[ipv].pv_move);
		/*
		Search the position
		*/
		for (ibook = 0; ibook < nbook[my_colour]; ibook++)
		{
			with = book[my_colour] + ibook;
			if ((with->book_tot_count[WHITE] == tot_count[WHITE]) and
			    (with->book_tot_count[BLACK] == tot_count[BLACK]) and
			    (with->book_total_key.hi == node->node_total_key.hi) and
			    (with->book_total_key.lo == node->node_total_key.lo)) break;
		}
		if (ibook < nbook[my_colour])
		{
			pv[ipv].pv_hit = with->book_hit;
			/*
			We ignore non-captures if we have a capture.
			*/
			if (has_capt and !(pv[ipv].pv_move & CAPT))
			{
				undo_move(my_colour, your_colour);
				continue;
			}
			/*
			If the move is a capture, reset ntemp and hit_max.
			*/
			if (pv[ipv].pv_move & CAPT)
			{
				if (!has_capt)
				{
					has_capt = TRUE;
					ntemp = 0;
					hit_max = 0;
				}
			}
			/*
			We choose the move with the highest hits.
			*/
			if (with->book_hit >= hit_max)
			{
				if (with->book_hit > hit_max)
				{
					hit_max = with->book_hit;
					ntemp = 0;
				}
				temp[ntemp] = ipv;
				ntemp++;
			}
		}
		else
			pv[ipv].pv_hit = 0;
		undo_move(my_colour, your_colour);
	}
	if (ntemp > 0)
	{
		my_printf("%sbook move(s)\n", colour_as_string[my_colour]);
		itemp = 0;
		for (ipv = 0; ipv < npv; ipv++)
		{
			if (pv[ipv].pv_hit > 0)
			{
				my_printf(" %s(%d)", pv[ipv].pv_string, pv[ipv].pv_hit);
				if ((itemp % 8) == 7) my_printf("\n");
				itemp++;
			}
		}
		my_printf("\n");
		my_printf("max hit(s)\n");
		for (itemp = 0; itemp < ntemp; itemp++)
		{
			my_printf(" %s", pv[temp[itemp]].pv_string);
			if (((itemp % 8) == 7) or (itemp == (ntemp - 1))) my_printf("\n");
		}
		return(temp[(int) (ulrand() % ntemp)]);
	}
	else
	{
		my_printf("%sis out of book!\n", colour_as_string[my_colour]);
		return(INVALID);
	}
}

/*
sort_pv sorts the pv moves by searching each move with a window of
-inf, inf and a depth of 1.
(not used yet).
*/

void sort_pv(int my_colour, int your_colour)
{
	int ipv, kpv;
	pv_t temp;
	
	clear_cache();
	node->node_pos[WHITE] = node->node_pos[BLACK] = 0;
	inode = 0;
	
	clock_begin = clock();
	time_used = 0;
	TimeLimit = 100;
	
	nodes = ngens = ndos = 0;
	nevals = nhits = nextends = 0;

	principle_variation = TRUE;
	for (ipv = 0; ipv < npv; ipv++)
	{
		do_move(my_colour, your_colour, pv[ipv].pv_move);
		pv[ipv].pv_score = alpha_beta(your_colour, my_colour, -SCORE_LIMIT, SCORE_LIMIT, 0);
		undo_move(my_colour, your_colour);
		my_printf("ts");
		analyze_cache(my_colour, your_colour, pv[ipv].pv_move);
		my_printf(" %d\n", pv[ipv].pv_score);
	}
	principle_variation = FALSE;
	BUG(null_move != FALSE)
	BUG(nthreats != 0)
	my_printf("time=%.2f stats=%ld/%ld/%ld/%ld/%ld/%ld\n",
		time_used, nodes, ngens, ndos, nevals, nhits, nextends);
	for (ipv = 0; ipv < (npv - 1); ipv++)
	{
		jpv = ipv;
		for (kpv = ipv + 1; kpv < npv; kpv++)
			if (pv[kpv].pv_score > pv[jpv].pv_score) jpv = kpv;
		if (jpv != ipv)
		{
			temp = pv[ipv];
			pv[ipv] = pv[jpv];
			pv[jpv] = temp;
		}
	}
	/*
	the best move is now the first move
	*/
	jpv = 0;
}

#define MARGIN (PAWN_SCORE / 2)

/*
do_pv tries to find the best move by iterative deepening,
aspiration search and minimal window search.
*/

void do_pv(int my_colour, int your_colour)
{
	int check;
	int depth;
	int temp_score, fh_score;
	int depth_next;
	int ipv;
	
	jpv = get_book_move(my_colour, your_colour);
	if (jpv != INVALID) return;
	
	clear_cache();
	check = attacked(king_list[my_colour]->board_where, your_colour);

	jpv = 0;
	pv[0].pv_score = node->node_mat[my_colour] - node->node_mat[your_colour];
	node->node_pos[WHITE] = node->node_pos[BLACK] = 0;
	inode = 0;
	
	clock_begin = clock();
	time_used = 0;
	
	ipv = 0;
	temp_score = SCORE_INVAL;
	/*
	Iterative deepening.
	*/
	for (depth = 1; depth <= DepthMax; depth++)
	{
		int alpha, beta;
		
		depth_next = (check ? depth : depth - 1);
		nodes = ngens = ndos = 0;
		nevals = nhits = nextends = 0;
		
		/*
		Aspiration sarch with half a pawn margin.
		*/

		alpha = pv[jpv].pv_score - MARGIN;
		beta = pv[jpv].pv_score + MARGIN;
		principle_variation = TRUE;
		do_move(my_colour, your_colour, pv[jpv].pv_move);
		temp_score = alpha_beta(your_colour, my_colour, alpha, beta, depth_next);
		undo_move(my_colour, your_colour);
		if (temp_score == SCORE_INVAL) goto label_final;
		do_move(my_colour, your_colour, pv[jpv].pv_move);
		
		/*
		Research fail low or fail high moves.
		*/
		
		if (temp_score <= alpha)
			temp_score = alpha_beta(your_colour, my_colour, -SCORE_LIMIT, temp_score, depth_next);
		else if (temp_score >= beta)
			temp_score = alpha_beta(your_colour, my_colour, temp_score, SCORE_LIMIT, depth_next);
		undo_move(my_colour, your_colour);
		if (temp_score == SCORE_INVAL) goto label_final;

		pv[jpv].pv_score = temp_score;
		my_printf("pv %d", depth);
		analyze_cache(my_colour, your_colour, pv[jpv].pv_move);
		my_printf(" %d\n", pv[jpv].pv_score);
		
		/*
		Search the remaining moves with a minimal window.
		*/
		
		principle_variation = FALSE;
		for (ipv = 0; ipv < npv; ipv++)
		{
			if (ipv == jpv) continue;
			do_move(my_colour, your_colour, pv[ipv].pv_move);
			temp_score = alpha_beta(your_colour, my_colour, pv[jpv].pv_score, pv[jpv].pv_score + 1, depth_next);
			undo_move(my_colour, your_colour);
			if (temp_score == SCORE_INVAL) goto label_final;
			
			/*
			Research if fail high.
			*/
			
			if (temp_score > pv[jpv].pv_score)
			{
				jpv = ipv;
				pv[jpv].pv_score = temp_score;
				my_printf("fh %d", depth);
				analyze_cache(my_colour, your_colour, pv[jpv].pv_move);
				my_printf(" %d\n", pv[jpv].pv_score);
				principle_variation = TRUE;
				do_move(my_colour, your_colour, pv[jpv].pv_move);
				fh_score = temp_score;
				temp_score = alpha_beta(your_colour, my_colour, temp_score, SCORE_LIMIT, depth_next);
				undo_move(my_colour, your_colour);
				if (temp_score == SCORE_INVAL) goto label_final;
				principle_variation = FALSE;
				
				/*
				If the reseach fails low, there was a null move failure (alas),
				a (depth >= cache_depth) failure (see alpha beta) or a hash problem.
				*/
				
				if (temp_score < fh_score)
					my_printf("null move failure or hash problem detected!\n");
				pv[jpv].pv_score = temp_score;
				my_printf("pv %d", depth);
				analyze_cache(my_colour, your_colour, pv[jpv].pv_move);
				my_printf(" %d\n", pv[jpv].pv_score);
			}
		}
		
		my_printf("pv %d", depth);
		analyze_cache(my_colour, your_colour, pv[jpv].pv_move);
		my_printf(" %d\n", pv[jpv].pv_score);
		my_printf("depth=%d score=%d time=%.2f stats=%ld/%ld/%ld/%ld/%ld/%ld\n",
			depth, pv[jpv].pv_score, time_used, nodes, ngens, ndos, nevals, nhits, nextends);
	}
	label_final:
	
	/*
	Only print this if time limit.
	*/
	
	if (temp_score == SCORE_INVAL)
	{
		my_printf("time limit depth=%d move=%d time=%.2f stats=%ld/%ld/%ld/%ld/%ld/%ld\n",
			depth, ipv, time_used, nodes, ngens, ndos, nevals, nhits, nextends);
	}
	BUG(null_move != FALSE)
	BUG(nthreats != 0)
}

/*
Initialization of some arrays.
*/

void init(void)
{
	int i, j, n;
	int icolour, ikind;
	FILE *f;
	
	/*
	The bit count arrays.
	*/
	
	MALLOC(bit_count, int, NBITS_SIZE)
	MALLOC(bit_low, int, NBITS_SIZE)
	MALLOC(bit_high, int, NBITS_SIZE)
	for (i = 0; i < NBITS_SIZE; i++)
	{
		n = 0;
		for (j = 0; j < NBITS; j++)
			if ((i & (1 << j)) != 0) n++;
		bit_count[i] = n;
		if (i == 0)
			bit_low[i] = bit_high[i] = INVALID;
		else
		{
			for (j = 0; j < NBITS; j++)
				if ((i & (1 << j)) != 0) break;
			bit_low[i] = j;
			for (j = NBITS - 1; j >= 0; j--)
				if ((i & (1 << j)) != 0) break;
			bit_high[i] = j;
		}
	}
	MALLOC(row, int, BOARD_MAX)
	MALLOC(column, int, BOARD_MAX)
	for (i = 0; i < BOARD_MAX; i++)
		row[i] = column[i] = -1;
	for (i = 0; i < 64; i++)
	{
		row[map[i]] = row_table[i];
		column[map[i]] = column_table[i];
	}
	MALLOC(nota, char *, BOARD_MAX)
	for (i = 0; i < BOARD_MAX; i++)
		nota[i] = nota_inval;
	for (i = 0; i < 64; i++)
		nota[map[i]] = nota_table[i];
	MALLOC(move_stack, move_t, MOVE_MAX)
	MALLOC(capt_stack, move_t, MOVE_MAX)
	move_top = move_stack;
	capt_top = capt_stack;
	MALLOC(node_stack, node_t, NODE_MAX)
	node_stack[0].node_reverse = FALSE;
	node_stack++;
	node = node_stack;
	inode = 0;
	
	my_printf("size of hash tables: %lu %lu %lu\n",
		sizeof(cache_t) * CACHE_MASK * 2,
		sizeof(score_cache_t) * SCORE_CACHE_MASK * 2,
		sizeof(quies_cache_t) * QUIES_CACHE_MASK * 2);
	MALLOC(cache[WHITE], cache_t, CACHE_MASK + 1)
	MALLOC(cache[BLACK], cache_t, CACHE_MASK + 1)
	MALLOC(score_cache[WHITE], score_cache_t, SCORE_CACHE_MASK + 1)
	MALLOC(score_cache[BLACK], score_cache_t, SCORE_CACHE_MASK + 1)
	MALLOC(quies_cache[WHITE], quies_cache_t, QUIES_CACHE_MASK + 1)
	MALLOC(quies_cache[BLACK], quies_cache_t, QUIES_CACHE_MASK + 1)

	/*
	The hash keys.
	*/
	
	for (icolour = WHITE; icolour <= BLACK; icolour++)
	{
		for (ikind = PAWN; ikind <= KING; ikind++)
		{
			MALLOC(hash_key[icolour][ikind], hash_key_t, BOARD_MAX)
			for (i = 0; i < 64; i++)
			{
				hash_key[icolour][ikind][map[i]].lo = ulrand();
				hash_key[icolour][ikind][map[i]].hi = ulrand();
			}
		}
	}
	MALLOC(ep_key, hash_key_t, BOARD_MAX)
	MALLOC(castle_key, hash_key_t, BOARD_MAX)
	for (i = 0; i < 64; i++)
	{
		ep_key[map[i]].lo = ulrand();
		ep_key[map[i]].hi = ulrand();
		castle_key[map[i]].lo = ulrand();
		castle_key[map[i]].hi = ulrand();
	}
	MALLOC(threat_key, hash_key_t, NODE_MAX)
	for (i = 0; i < NODE_MAX; i++)
	{
		threat_key[i].lo = ulrand();
		threat_key[i].hi = ulrand();
	}
	
	/*
	Some more tables.
	*/
	
	MALLOC(king_alone_grad, int, BOARD_MAX)
	MALLOC(white_free_pawn_grad, int, BOARD_MAX)
	MALLOC(black_free_pawn_grad, int, BOARD_MAX)
	free_pawn_grad[WHITE] = white_free_pawn_grad;
	free_pawn_grad[BLACK] = black_free_pawn_grad;

	MALLOC(white_pawn_visit, int, BOARD_MAX)
	MALLOC(black_pawn_visit, int, BOARD_MAX)
	pawn_visit[WHITE] = white_pawn_visit;
	pawn_visit[BLACK] = black_pawn_visit;

	MALLOC(white_knight_visit, int, BOARD_MAX)
	MALLOC(black_knight_visit, int, BOARD_MAX)
	knight_visit[WHITE] = white_knight_visit;
	knight_visit[BLACK] = black_knight_visit;

	MALLOC(pawn_board, int, BOARD_MAX)
	for (i = 0; i < BOARD_MAX; i++)
	{
		king_alone_grad[i] = 0;
		white_free_pawn_grad[i] = black_free_pawn_grad[i] = 0;
		white_pawn_visit[i] = black_pawn_visit[i] = 0;
		white_knight_visit[i] = black_knight_visit[i] = 0;
	}
	for (i = 0; i < 64; i++)
	{
		king_alone_grad[map[i]] = king_alone_grad_table[i];
		white_free_pawn_grad[map[i]] = free_pawn_grad_table[i];
		black_free_pawn_grad[mirror_map[i]] = free_pawn_grad_table[i];
		white_pawn_visit[map[i]] = pawn_visit_table[i];
		black_pawn_visit[mirror_map[i]] = pawn_visit_table[i];
		white_knight_visit[map[i]] = knight_visit_table[i];
		black_knight_visit[mirror_map[i]] = knight_visit_table[i];
	}
	
	/*
	Scale the free pawn gradient such that a pawn on rank 7 is worth a pawn
	*/
	
	for (i = 0; i < 64; i++)
	{
		white_free_pawn_grad[map[i]] *= 3;
		black_free_pawn_grad[map[i]] *= 3;
	}
	MALLOC(strength_board, int, BOARD_MAX)
	MALLOC(capt_strength_table, int, NPVS_MAX)
	MALLOC(move_strength_table, int, NPVS_MAX)
	MALLOC(attack_board, int, BOARD_MAX)
	
	/*
	Read the opening book.
	*/
	
	if (UseBook)
	{
		for (icolour = WHITE; icolour <= BLACK; icolour++)
		{
			f = fopen(book_name[icolour], "rb");
			if (f != NULL)
			{
				BUG(fread(&(nbook[icolour]), sizeof(long), 1, f) != 1)
				MALLOC(book[icolour], book_t, nbook[icolour])
				BUG(fread(book[icolour], sizeof(book_t), nbook[icolour], f) != nbook[icolour])
				my_printf("loaded %sbook size=%ld\n",
					colour_as_string[icolour], nbook[icolour]);
			}
		}
	}
}
